import socket
import os
import glob
import subprocess
import time
import zmq
import numpy as np

__version__ = '0.3.1'

class eCubeStream:

    VERSTR = b"ECZQ01"

    NUM_CH_ANALOG = 32
    NUM_VCH_DIGITAL = 64
    NUM_CH_HSBANK = 640

    CONST_HS_GAIN = 1.907348633e-7
    CONST_ANALOG_GAIN = 3.0517578125e-4

    dtypedict = {
    'Headstages': np.int16,
    'AnalogPanel': np.int16,
    'DigitalPanel': np.uint64
    }

    def __init__(self, sources=None, asfloat=False, snaddress=None, ctladdress='127.0.0.1', ctlport=49686, dataport=49676, autoshutdown=True, debug=False):
        '''Initializes a pyeCubeStream object. This is the default usage of this wrapper class.'''
        self.autoshutdown = autoshutdown
        self.proc = None
        self.debug = debug
        self.retryconnect = False

        self.snaddress = snaddress

        self.__testconnection(endpoint=(ctladdress, ctlport))

        self.ctladdr = 'tcp://'+ctladdress+':'+str(ctlport)
        self.dataaddr = 'tcp://'+ctladdress+':'+str(dataport)
        self.ctx = zmq.Context.instance()
        self.ctlsock = self.ctx.socket(zmq.REQ)
        self.datasock = self.ctx.socket(zmq.SUB)

        self.ctlconnect()

        self.opts_digitalchans = None
        self.opts_digitalchanfilter = None
        self.opts_headstagefloat = None
        self.opts_analogfloat = None

        self.isstreaming = False

        if sources is not None:
            self.add(sources, asfloat)

    def __testconnection(self, endpoint):
        '''Tests the ctladdr endpoint to see if can connect; if not, try to start servernode-control'''
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.25)
            if s.connect_ex(endpoint) != 0:
                address = endpoint[0]
                ctlport = endpoint[1]
                if self.retryconnect or address not in ('127.0.0.1', 'localhost'):
                    raise RuntimeError('Cannot connect to remote address ' + str(address) + ':' + str(ctlport))
                else:
                    self.__startctlhost(endpoint)

    def __startctlhost(self, endpoint):
        '''Start servernode-control by looking for it in import path, ../ from import path, or cwd'''
        cwdir = os.getcwd()
        libdir = os.path.dirname(os.path.realpath(__file__))
        libdirp = os.path.split(libdir)[0]

        exelist = glob.glob(os.path.join(cwdir, "servernode-control*"))
        exelist.extend(glob.glob(os.path.join(libdir, "servernode-control*")))
        exelist.extend(glob.glob(os.path.join(libdirp, "servernode-control*")))

        if len(exelist) < 1:
            raise RuntimeError('Cannot connect to remote address, cannot find servernode-control')
        elif len(exelist) > 1:
            raise RuntimeError('Cannot connect to remote address, multiple servernode-control found')

        launchcmd = [exelist[0]]

        if self.snaddress is not None and isinstance(self.snaddress, str):
            launchcmd.append(self.snaddress)

        if self.debug is True:
            self.proc = subprocess.Popen(launchcmd)
        else:
            self.proc = subprocess.Popen(launchcmd, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        time.sleep(3)

        if self.proc.poll() is not None:
            raise RuntimeError('Error from servernode-control process detected')

        self.retryconnect = True
        self.__testconnection(endpoint)

    def ctlconnect(self):
        '''Connect the zmq interface'''
        if self.debug:
            print("Connecting eCubeZMQCtl to " + self.ctladdr)
        self.ctlsock.connect(self.ctladdr)

    def ctldisconnect(self):
        '''Disconnect the zmq interface'''
        if self.debug:
            print("Disconnecting eCubeZMQCtl from " + self.ctladdr)
        self.ctlsock.disconnect(self.ctladdr)

    def listavailable(self):
        '''Returns all data sources currently available for subscribing, from the upstream eCube or ServerNode.'''
        self.ctlsock.send_multipart([self.VERSTR, b"LISTAVAIL"])
        reply = self.ctlsock.recv_multipart()

        if reply[0] == b"OKAVAIL":
            availchans = np.frombuffer(reply[1], dtype='<i4')

            if self.debug is True:
                print("Channels available:")
                for hsnum, hscnt in zip(range(availchans.size), availchans):
                    if hscnt > 0:
                        print("  HS {}: {}".format(hsnum+1, hscnt))
                print("  AnalogPanel\n  DigitalPanel")
        else:
            raise ValueError("listavailable() reply invalid")
        return (availchans, self.NUM_CH_ANALOG, self.NUM_VCH_DIGITAL)

    def listadded(self):
        '''Returns all data sources currently in subscription.'''
        self.ctlsock.send_multipart([self.VERSTR, b"LISTADDED"])
        reply = self.ctlsock.recv_multipart()

        if reply[0] == b"OKADDLIST" and len(reply) == 5:

            hschans = np.ndarray((0), dtype='<u4')
            achans = np.ndarray((0), dtype='<u4')
            dchans = np.ndarray((0), dtype='<u4')

            if len(reply[1]) >= 4:
                hschanlinear = np.frombuffer(reply[1], dtype='<u4')
                hschansarr = np.asarray(list(map(
                    lambda ch_i: divmod(ch_i, self.NUM_CH_HSBANK),
                    hschanlinear)), dtype='<u4') + 1
                hschans = ((hschansarr[:, 0]).squeeze(), (hschansarr[:, 1]).squeeze())
                if self.debug is True:
                    print("Headstage channels added:")
                    print(hschans)

            if len(reply[2]) >= 4:
                achans = np.frombuffer(reply[2], dtype='<u4') + 1
                if self.debug is True:
                    print("AnalogPanel channels added:")
                    print(achans)

            if len(reply[3]) >= 4:
                dchans = np.frombuffer(reply[3], dtype='<u4') + 1
                if self.debug is True:
                    print("DigitalPanel channels added:")
                    print(dchans)

            return (hschans, achans, dchans)

        else:
            raise ValueError("listadded() reply invalid")

    def __sourcevalidate(self, source):
        '''Validate the format of a single source in .add() or .remove() operations and return readable suggestions.'''
        if type(source) is list:
            if len(source) < 1:
                raise ValueError("Source list cannot be empty")

            if all([self.__sourcevalidate(sourceitem) for sourceitem in source]):
                return True

            return False

        elif type(source) is not tuple:
            raise TypeError("Each source must be a tuple e.g. ('Headstages', 2, (1,8)), ('DigitalPanel',)")

        if len(source) < 1:
            raise ValueError("Each source must be a tuple of size 1-3")

        if type(source[0]) is not str:
            raise TypeError("The first tuple element of each source must be 'Headstages', 'AnalogPanel', or 'DigitalPanel'")

        if source[0] == "Headstages":
            if len(source) < 2 or len(source) > 3:
                raise ValueError("Source must specify a Headstage#, and optional channel range tuple ()")
            if type(source[1]) is not int or source[1] < 1 or source[1] > 10:
                raise ValueError("Headstage# must be an integer between 1 - 10")
            if len(source) == 3:
                if type(source[2]) is not tuple or len(source[2]) != 2:
                    raise TypeError("Headstage channel range must be a tuple (startch, stopch)")
                if source[2][0] < 1 or source[2][0] > self.NUM_CH_HSBANK or source[2][1] < 1 or source[2][1] > self.NUM_CH_HSBANK:
                    raise ValueError("Headstage channels must be between 1 - " + str(self.NUM_CH_HSBANK))
                if source[2][1] < source[2][0]:
                    raise ValueError("Headstage channel (startch, stopch) range invalid")
            return True

        elif source[0] == "AnalogPanel":
            if len(source) < 1 or len(source) > 2:
                raise TypeError("AnalogPanel must be a tuple ('AnalogPanel', (startch, stopch))")
            if len(source) == 2:
                if type(source[1]) is not tuple or len(source[1]) != 2:
                    raise TypeError("AnalogPanel channel range must be a tuple (startch, stopch)")
                if source[1][0] < 1 or source[1][0] > self.NUM_CH_ANALOG or source[1][1] < 1 or source[1][1] > self.NUM_CH_ANALOG:
                    raise ValueError("AnalogPanel channels must be between 1 - " + str(self.NUM_CH_ANALOG))
                if source[1][1] < source[1][0]:
                    raise ValueError("AnalogPanel channel (startch, stopch) range invalid")
            return True
        elif source[0] == "DigitalPanel":
            if len(source) < 1 or len(source) > 2:
                raise TypeError("DigitalPanel must be a tuple ('DigitalPanel', (startch, stopch))")
            if len(source) == 2:
                if type(source[1]) is not tuple or len(source[1]) != 2:
                    raise TypeError("DigitalPanel channel range must be a tuple (startch, stopch)")
                if source[1][0] < 1 or source[1][0] > self.NUM_VCH_DIGITAL or source[1][1] < 1 or source[1][1] > self.NUM_VCH_DIGITAL:
                    raise ValueError("DigitalPanel channels must be between 1 - " + str(self.NUM_VCH_DIGITAL))
                if source[1][1] < source[1][0]:
                    raise ValueError("DigitalPanel channel (startch, stopch) range invalid")
            return True

        elif source[0] == "DigitalPanelAsChans":
            if len(source) < 1 or len(source) > 2:
                raise TypeError("DigitalPanelAsChans must be a tuple ('DigitalPanelAsChans', (startch, stopch))")
            if len(source) == 2:
                if type(source[1]) is not tuple or len(source[1]) != 2:
                    raise TypeError("DigitalPanelAsChans channel range must be a tuple (startch, stopch)")
                if source[1][0] < 1 or source[1][0] > self.NUM_VCH_DIGITAL or source[1][1] < 1 or source[1][1] > self.NUM_VCH_DIGITAL:
                    raise ValueError("DigitalPanelAsChans channels must be between 1 - " + str(self.NUM_VCH_DIGITAL))
                if source[1][1] < source[1][0]:
                    raise ValueError("DigitalPanelAsChans channel (startch, stopch) range invalid")
            return True
        else:
            print(source)
            raise ValueError("Unknown source type")

    def add(self, sources, asfloat=False):
        '''Adds data sources to the current subscription. Iterate through if source is a list.'''
        if not self.__sourcevalidate(sources):
            raise ValueError("Invalid sources")

        if type(sources) is list:
            for source in sources:
                self.__addsingle(source, asfloat)
        else:
            self.__addsingle(sources, asfloat)

    def __addsingle(self, source, asfloat=False):
        '''Called by .add() to actually add sources.'''
        if source[0] == "Headstages":
            if len(source) == 2:
                self.ctlsock.send_multipart([self.VERSTR, b"ADD", b"H",
                    source[1].to_bytes(4, 'little')])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))

            elif len(source) == 3:
                self.ctlsock.send_multipart(
                    [self.VERSTR, b"ADD", b"H",
                    source[1].to_bytes(4, 'little'),
                    source[2][0].to_bytes(4, 'little'),
                    source[2][1].to_bytes(4, 'little')])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))
            else:
                raise ValueError("Unknown Headstage add command")

            if asfloat is True:
                self.opts_headstagefloat = True

        elif source[0] == "AnalogPanel":
            if len(source) == 1:
                self.ctlsock.send_multipart([self.VERSTR, b"ADD", b"A"])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))

            elif len(source) == 2:
                self.ctlsock.send_multipart(
                    [self.VERSTR, b"ADD", b"A",
                    source[1][0].to_bytes(4, 'little'),
                    source[1][1].to_bytes(4, 'little')])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))
            else:
                raise ValueError("Unknown AnalogPanel add command")

            if asfloat is True:
                self.opts_analogfloat = True

        elif source[0] == "DigitalPanel" or source[0] == "DigitalPanelAsChans":
            if (self.opts_digitalchans is not None) and ((source[0] == "DigitalPanel" and self.opts_digitalchans is not False) or (source[0] == "DigitalPanelAsChans" and self.opts_digitalchans is not True)):
                raise RuntimeError("DigitalPanel already subscribed. Please remove before changing subscription type.")

            if len(source) == 1:
                self.ctlsock.send_multipart([self.VERSTR, b"ADD", b"D"])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))

            elif len(source) == 2:
                self.ctlsock.send_multipart(
                    [self.VERSTR, b"ADD", b"D",
                    source[1][0].to_bytes(4, 'little'),
                    source[1][1].to_bytes(4, 'little')])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKADD":
                    raise ValueError(reply[0].decode('utf-8'))
            else:
                raise ValueError("Unknown DigitalPanel add command")

            if source[0] == "DigitalPanelAsChans":
                self.opts_digitalchans = True

                if len(source) == 1:
                    self.opts_digitalchanfilter = None
                if len(source) == 2:
                    if self.opts_digitalchanfilter is None:
                        self.opts_digitalchanfilter = list(range(source[1][0]-1, source[1][1]))
                    elif type(self.opts_digitalchanfilter) is list:
                        self.opts_digitalchanfilter = list(set(self.opts_digitalchanfilter).union(set(range(source[1][0]-1, source[1][1]))))
                    else:
                        raise TypeError("Digital channels python API is the wrong type")
                else:
                    raise ValueError("Unknown DigitalPanelAsChans add command")
            else:
                self.opts_digitalchans = False

        else:
            raise ValueError("Unknown add command")

    def remove(self, sources):
        '''Removes data sources from the current subscription. Iterate through if source is a list.'''
        if not self.__sourcevalidate(sources):
            raise ValueError("Invalid sources")

        if type(sources) is list:
            for source in sources:
                self.__removesingle(source)
        else:
            self.__removesingle(sources)

    def __removesingle(self, source):
        '''Called by .remove() to actually remove sources from subscription.'''
        if source[0] == "Headstages":
            if len(source) == 2:
                self.ctlsock.send_multipart([self.VERSTR, b"REMOVE", b"H",
                    source[1].to_bytes(4, 'little')])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKREMOVE":
                    raise ValueError(reply[0].decode('utf-8'))
                self.opts_headstagefloat = None

            elif len(source) == 3:
                raise ValueError("Removal operations on whole-headstage only, e.g. ('Headstages', 2)")

            else:
                raise ValueError("Unknown headstage remove command")

        elif source[0] == "AnalogPanel":
            if len(source) == 1:
                self.ctlsock.send_multipart([self.VERSTR, b"REMOVE", b"A"])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKREMOVE":
                    raise ValueError(reply[0].decode('utf-8'))
                self.opts_analogfloat = None

            elif len(source) == 2:
                raise ValueError("Removal operations on whole-AnalogPanel only, e.g. ('AnalogPanel',)")

            else:
                raise ValueError("Unknown AnalogPanel remove command")

        elif source[0] == "DigitalPanel" or source[0] == "DigitalPanelAsChans":
            if len(source) == 1:
                self.ctlsock.send_multipart([self.VERSTR, b"REMOVE", b"D"])
                reply = self.ctlsock.recv_multipart()
                if len(reply) == 1 and reply[0] != b"OKREMOVE":
                    raise ValueError(reply[0].decode('utf-8'))
                self.opts_digitalchans = None
                self.opts_digitalchanfilter = None

            elif len(source) == 2:
                raise ValueError("Removal operations on whole-DigitalPanel only, e.g. ('DigitalPanel',)")

            else:
                raise ValueError("Unknown DigitalPanel remove command")

        else:
            raise ValueError("Unknown remove command")

    def __closesession(self):
        self.ctlsock.send_multipart([self.VERSTR, b"CLOSESESSION"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKCLOSE":
            raise RuntimeError(reply[0].decode('utf-8'))

    def remotesave(self, sessionname):
        '''Start a remote data recording on the host running servernode software, based on the currently subscribed set of channels. The recordings will be saved into a newly created directory matching the method parameter sessionname.'''
        if self.__hassessions() is not True:
            raise RuntimeError("Upstream connection is not ServerNode and cannot perform remote saving")

        if type(sessionname) is not str:
            raise RuntimeError("sessionname must be a valid string")
        sessionbytes = sessionname.encode('utf-8')

        self.ctlsock.send_multipart([self.VERSTR, b"CREATESESSION", sessionbytes])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKCREATE":
            raise RuntimeError(reply[0].decode('utf-8'))

        self.ctlsock.send_multipart([self.VERSTR, b"RECORDSESSION", b"START"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKSTART":
            raise RuntimeError(reply[0].decode('utf-8'))

        self.__closesession()

    def listremotesessions(self):
        '''Lists all currently created remote sessions running on the host running servernode software.'''
        if self.__hassessions() is not True:
            raise RuntimeError("Upstream connection is not ServerNode and cannot perform remote saving")

        self.ctlsock.send_multipart([self.VERSTR, b"LISTSESSIONS"])
        reply = self.ctlsock.recv_multipart()

        if reply[0] == b"OKLIST":
            if self.debug is True:
                print("Remove saving sessions open:")
                for sess in reply[1:]:
                    print("  " + sess.decode('utf-8'))
            return [sess.decode('utf-8') for sess in reply[1:]]
        else:
            raise ValueError("listremotesessions() reply invalid")

    def remotestopsave(self, sessionname):
        '''Stop an active remote data recording on the host running servernode software.'''
        if self.__hassessions() is not True:
            raise RuntimeError("Upstream connection is not ServerNode and cannot perform remote saving")

        if type(sessionname) is not str:
            raise RuntimeError("sessionname must be a valid string")
        sessionbytes = sessionname.encode('utf-8')

        self.ctlsock.send_multipart([self.VERSTR, b"OPENSESSION", sessionbytes])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKOPEN":
            raise RuntimeError(reply[0].decode('utf-8'))

        self.ctlsock.send_multipart([self.VERSTR, b"RECORDSESSION", b"STOP"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKSTOP":
            raise RuntimeError(reply[0].decode('utf-8'))

        self.ctlsock.send_multipart([self.VERSTR, b"DESTROYSESSION"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKDESTROY":
            raise RuntimeError(reply[0].decode('utf-8'))

    def start(self):
        '''Starts the concurrent streaming of subscribed data sources, and makes available obtaining streaming data via .get().'''
        self.datasock.connect(self.dataaddr)
        self.datasock.setsockopt_string(zmq.SUBSCRIBE, '')

        self.ctlsock.send_multipart([self.VERSTR, b"START"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKSTART":
            raise RuntimeError(reply[0].decode('utf-8'))

        self.isstreaming = True
        return True

    def get(self):
        '''Obtains the latest available data packet from the stream.'''
        if self.isstreaming is not True:
            raise RuntimeError("get() must be called while streaming by issuing start()")

        datamsg = self.datasock.recv_multipart()
        samples = int.from_bytes(datamsg[2], byteorder='little', signed=True)

        sourcetype = datamsg[0].decode('utf-8')
        if sourcetype not in self.dtypedict.keys():
            raise TypeError('Invalid sourcetype ' + sourcetype + ' detected.')

        if sourcetype == 'DigitalPanel':
            channels = 1
        else:
            channels = int.from_bytes(datamsg[3], byteorder='little', signed=False)

        if len(datamsg[4]) != channels * samples * np.dtype(self.dtypedict[sourcetype]).itemsize:
            raise ValueError("Error: data packet size {} did not match {} samples of {} channels".format(
                len(datamsg[4]), samples, channels))

        timestamp = int.from_bytes(datamsg[1], byteorder='little', signed=False)

        if sourcetype == 'DigitalPanel' and self.opts_digitalchans is True:
            data = np.frombuffer(datamsg[4], dtype='<u1')
            data = np.unpackbits(data, bitorder='little')
            channels = self.NUM_VCH_DIGITAL
        else:
            data = np.frombuffer(datamsg[4], dtype=self.dtypedict[sourcetype])

        data = data.reshape((samples, channels))

        if self.opts_headstagefloat is True and sourcetype == 'Headstages':
            data = data.astype(np.single) * self.CONST_HS_GAIN

        if self.opts_analogfloat is True and sourcetype == 'AnalogPanel':
            data = data.astype(np.single) * self.CONST_ANALOG_GAIN

        if sourcetype == 'DigitalPanel' and self.opts_digitalchans is True and self.opts_digitalchanfilter is not None:
            data = data[:, self.opts_digitalchanfilter]
            sourcetype = 'DigitalPanelAsChans'

        return (timestamp, sourcetype, data)

    def stop(self):
        '''Stops the concurrent streaming of subscribed data sources.'''
        self.ctlsock.send_multipart([self.VERSTR, b"STOP"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKSTOP":
            raise ValueError(reply[0].decode('utf-8'))

        self.datasock.disconnect(self.dataaddr)

        self.isstreaming = False
        return True

    def resetheadstage(self, hsnum):
        '''Sends headstage reset command.'''

        if not isinstance(hsnum, int) or hsnum < 1 or hsnum > 10:
            raise ValueError("Headstage# must be an integer between 1 - 10")

        self.ctlsock.send_multipart([self.VERSTR, b"RESETHS",
            hsnum.to_bytes(4, 'little')])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] != b"OKRESETHS":
            raise ValueError(reply[0].decode('utf-8'))

        return True

    def __hassessions(self):
        self.ctlsock.send_multipart([self.VERSTR, b"HASSESSIONS"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] == b"OKSESSIONS":
            return True
        return False

    def __isstreaming(self):
        self.ctlsock.send_multipart([self.VERSTR, b"ISSTREAMING"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] == b"OKSTREAMING":
            return True
        return False

    def __issession(self):
        self.ctlsock.send_multipart([self.VERSTR, b"ISSESSION"])
        reply = self.ctlsock.recv_multipart()
        if len(reply) == 1 and reply[0] == b"OKINSESSION":
            return True
        return False

    def __del__(self):
        '''Cleans up on shutdown.'''
        if self.autoshutdown:

            if hasattr(self, 'ctlsock'):
                if self.__isstreaming():
                    self.stop()

                if self.__issession():
                    self.__closesession()

                self.ctldisconnect()

            if self.proc is not None:
                try:
                    outs, errs = self.proc.communicate(input="quit\n", timeout=2)
                except:
                    self.proc.kill()
                    outs, errs = self.proc.communicate()



import copy
import socket
hostname = socket.gethostname()

# Defaults
rig_name = hostname
optitrack_address = None
optitrack_save_path = None
optitrack_sync_dch = 0
window_size = (1920, 1080)
screen_dist = 45
screen_half_height = 12
default_db = 'local'
secret_dbnames = []
reward_address = '/dev/rewardsystem'
force_sensor_address = '/dev/forcesensor'
reward_digital_pin = 12
sync_events = dict(
    EXP_START               = 0x1,
    TRIAL_START             = 0x2,
    TARGET_ON               = 0x10,
    TARGET_OFF              = 0x20,
    REWARD                  = 0x30,
    PARTIAL_REWARD          = 0x31,
    HOLD_PENALTY            = 0x40,
    TIMEOUT_PENALTY         = 0x41,
    DELAY_PENALTY           = 0x42,
    FIXATION_PENALTY        = 0x43,
    OTHER_PENALTY           = 0x4f,
    CURSOR_ENTER_TARGET     = 0x50,
    CURSOR_LEAVE_TARGET     = 0x60,
    CUE                     = 0x70,
    PAUSE_START             = 0x80,
    PAUSE_END               = 0x81,
    FIXATION                = 0x90,
    EYE_TARGET_ON           = 0xa0,
    EYE_TARGET_OFF          = 0xb0,
    TIME_ZERO               = 0xee,
    TRIAL_END               = 0xef,
    PAUSE                   = 0xfe,
    EXP_END                 = 0xff,    # For ease of implementation, the last event must be the highest possible value
)

hdf_sync_params = dict(
    sync_protocol = 'hdf',
    sync_protocol_version = 0,
    sync_pulse_width = 0.,
    event_sync_dict = sync_events,
    event_sync_max_data = 0xd,
    event_sync_data_shift = 0,
)
nidaq_sync_params = copy.copy(hdf_sync_params)
nidaq_sync_params.update(dict(
    sync_protocol = 'rig1',
    sync_protocol_version = 10,
    sync_pulse_width = 0.003,
    event_sync_mask = 0xffffff,
    event_sync_dch = list(range(16,24)),
    screen_sync_pin = 8,
    screen_sync_dch = 24,
    screen_measure_dch = [5],
    screen_measure_ach = [5],
    reward_measure_ach = [0],
    right_eye_ach = [8, 9],
    left_eye_ach = [10, 11],
    recording_pin = 9,
    recording_dch = 25,
))
rig1_sync_params_arduino = copy.copy(nidaq_sync_params)
rig1_sync_params_arduino.update(dict(
    sync_protocol = 'rig1_arduino',
    sync_protocol_version = 16,
    baudrate = 115200,
    event_sync_mask = 0xfffffc,
    event_sync_data_shift = 2,
    event_sync_dch = list(range(31,39)),
    screen_sync_pin = 10,
    screen_sync_dch = 39,
    recording_pin = 11,
    recording_dch = 40,
))
rig2_sync_params_arduino = copy.copy(rig1_sync_params_arduino)
rig2_sync_params_arduino.update(dict(
    sync_protocol = 'rig2',
    sync_protocol_version = 15,
    event_sync_dch = [41,42,43,44,45,48,49,50],
    screen_sync_dch = 51,
    screen_measure_dch = [6],
    screen_measure_ach = [6],
    reward_measure_ach = [1],
    right_eye_ach = [26, 27],
    left_eye_ach = [24, 25],
    recording_dch = 52,
))
arduino_sync_params = None
human_sync_params_arduino = copy.copy(hdf_sync_params)
human_sync_params_arduino.update(dict(
    sync_protocol = 'human_arduino',
    sync_protocol_version = 17,    
    baudrate = 115200,
    sync_pulse_width = 0.003,
    event_sync_mask = 0x400, # For now just send the sync pulse on the first digital channel
    event_sync_data_shift = 11,
    screen_sync_pin = 10,
    screen_sync_dch = 0,
    recording_pin = 9,
))

# Rig-specific defaults
if hostname == 'pagaiisland2':
    optitrack_address = '10.155.206.1'
    optitrack_save_path = "C:/Users/Orsborn Lab/Documents"
    window_size = (2560, 1440)
    screen_dist = 28
    screen_half_height = 10.75
    default_db = 'booth'
    secret_dbnames = ['booth', 'test']
    arduino_sync_params = rig1_sync_params_arduino
elif hostname == 'siberut-bmi':
    optitrack_address = '10.155.204.10'
    optitrack_save_path = "C:/Users/aolab/Documents"
    optitrack_sync_dch = 53 # 0-index
    screen_dist = 28
    screen_half_height = 10.25
    default_db = 'booth'
    secret_dbnames = ['booth', 'test']
    reward_digital_pin = 2
    arduino_sync_params = rig2_sync_params_arduino
elif hostname == 'booted-server':
    screen_half_height = 5
    default_db = 'tablet'
    secret_dbnames = ['tablet']
elif hostname == 'human-bmi':
    optitrack_address = '128.95.215.191'
    optitrack_save_path = "D:\optitrack_data"
    optitrack_sync_dch = 53 # 0-index
    window_size = (1920, 1080)
    screen_dist = 80
    screen_half_height = 16
    default_db = 'human'
    secret_dbnames = ['human']
    arduino_sync_params = human_sync_params_arduino
elif hostname in ['moor', 'crab-eating', 'ecube']:
    default_db = 'booth'
    secret_dbnames = [
        'human',
        'booth',
        'tablet',
        'test',
    ]

# Organize the settings
rig_settings = {
    'name': rig_name,
}

reward = {
    'address': reward_address,
    'digital_pin': reward_digital_pin,
}

optitrack = {
    'offset': [0, -60, -30], # optitrack cm [forward, up, right]
    'scale': 1 ,# optitrack cm --> screen cm
    'address': optitrack_address,
    'save_path': optitrack_save_path,
    'sync_dch': optitrack_sync_dch,
}

window = {
    'window_size': window_size,
    'screen_dist': screen_dist,
    'screen_half_height': screen_half_height,
}

db = {
    'secret_dbnames': secret_dbnames,
    'enable_celery': False,
    'default_db': default_db,
}



'''
Generic data source module. Sources run in separate processes and continuously collect/save
data and interact with the main process through the methods here.
'''
import time
import inspect
import traceback
import multiprocessing as mp
from multiprocessing import sharedctypes as shm

import hdfwriter
import numpy as np

from riglib.hdfwriter import supp_hdf
from .mp_proxy import FuncProxy, RPCProcess

from . import sink

class DataSourceSystem(object):
    '''
    Abstract base class for use with the generic DataSource infrastructure. Requirements:
        1) the class must have an attribute named 'dtype' which represents 
           the data type of the source data. The datatype *cannot* change!
        2) the class must have an attribute named 'update_freq' which specifies 
           the frequency at which new data samples will be ready for retrieval.
        3) 'start' method--no arguments
        4) 'stop' method--no arguments
        5) 'get' method--should return a single output argument
    '''
    dtype = np.dtype([])
    update_freq = 1
    def start(self):
        '''
        Initialization for the source
        '''
        pass

    def stop(self):
        '''
        Code to run when the data source is to be stopped
        '''
        pass

    def get(self):
        '''
        Retrieve the current data available from the source. 
        '''
        pass


class DataSource(RPCProcess):
    def __init__(self, source, bufferlen=10, name=None, send_data_to_sink_manager=True, source_kwargs=dict(), **kwargs):
        '''
        Parameters
        ----------
        source: class compatible with DataSourceSystem
            Class to be instantiated as the "system" with changing data values. 
        bufferlen: float
            Number of seconds long to make the ringbuffer. Seconds are converted to number 
            of samples based on the 'update_freq' attribute of the source
        name: string, optional, default=None
            Name of the sink, i.e., HDF table. If one is not provided, it will be inferred based
            on the name of the source module
        send_data_to_sink_manager: boolean, optional, default=True
            Flag to indicate whether data should be saved to a sink (e.g., HDF file)
        kwargs: optional keyword arguments
            Passed to the source during object construction if any are specified

        Returns
        -------
        DataSource instance
        '''
        super(DataSource, self).__init__(**kwargs)
        if name is not None:
            self.name = name
        else:
            self.name = source.__module__.split('.')[-1]
        self.filter = None
        self.target_class = self.source = source
        self.target_kwargs = self.source_kwargs = source_kwargs #kwargs
        self.bufferlen = bufferlen
        self.max_len = bufferlen * int(self.source.update_freq)
        self.slice_size = self.source.dtype.itemsize
        
        self.lock = mp.Lock()
        self.idx = shm.RawValue('l', 0)
        self.data = shm.RawArray('c', self.max_len * self.slice_size)
        # self.pipe, self._pipe = mp.Pipe()
        # self.cmd_event = mp.Event()
        # self.status = mp.Value('b', 1)
        self.stream = mp.Event()
        self.last_idx = 0
        self.streaming = False


        # in DataSource.run, there is a call to "self.sinks.send(...)",
        # but if the DataSource was never registered with the sink manager,
        # then this line results in unnecessary IPC
        # so, set send_data_to_sink_manager to False if you want to avoid this
        self.send_data_to_sink_manager = send_data_to_sink_manager

    def target_constr(self):
        try:
            self.target = self.target_class(**self.target_kwargs)
            self.target.start()
        except Exception as e:
            print("source.DataSource.run: unable to start source!")
            print(e)

            import io
            err = io.StringIO()
            self.log_error(err, mode='a')
            err.seek(0)

            self.status.value = -1

        self.streaming = True

    def loop_task(self):
            size = self.slice_size

            if self.stream.is_set():
                self.stream.clear()
                self.streaming = not self.streaming
                if self.streaming:
                    self.idx.value = 0
                    self.target.start()
                else:
                    self.target.stop()

            if self.streaming:
                data = self.target.get()
                if self.send_data_to_sink_manager:
                    sink_manager = sink.SinkManager.get_instance()
                    sink_manager.send(self.name, data)
                if data is not None:
                    try:
                        self.lock.acquire()
                        i = self.idx.value % self.max_len
                        self.data[i*size:(i+1)*size] = np.array(data).tobytes()
                        self.idx.value += 1
                        self.lock.release()
                    except Exception as e:
                        print("source.DataSource.run, exception saving data to ring buffer")
                        print(e)
            else:
                time.sleep(.001)        

    def target_destr(self, ret_status, msg):
        # stop the system once self.status.value has been set to a negative number
        self.target.stop()

    def get(self, all=False, **kwargs):
        '''
        Retreive data from the remote process

        Parameters
        ----------
        all : boolean, optional, default=False
            If true, returns all the data currently available. Since a finite buffer is used, 
            this is NOT the same as all the data observed. (see 'bufferlen' in __init__ for buffer size)
        kwargs : optional kwargs 
            To be passed to self.filter, if it is listed

        Returns
        -------
        np.recarray 
            Datatype of record array is the dtype of the DataSourceSystem
        '''
        if self.status.value <= 0:
            raise Exception('\n\nError starting datasource: %s\n\n' % self.name)
            
        self.lock.acquire()
        i = (self.idx.value % self.max_len) * self.slice_size
        if all:
            if self.idx.value < self.max_len:
                data = self.data[:i]
            else:
                data = self.data[i:]+self.data[:i]
        else:
            mlen = min((self.idx.value - self.last_idx), self.max_len)
            last = ((self.idx.value - mlen) % self.max_len) * self.slice_size
            if last > i:
                data = self.data[last:] + self.data[:i]
            else:
                data = self.data[last:i]
            
        self.last_idx = self.idx.value
        self.lock.release()
        try:
            data = np.frombuffer(data, dtype=self.source.dtype)
        except:
            print("can't get fromstring...")

        if self.filter is not None:
            return self.filter(data, **kwargs)
        return data

    def pause(self):
        '''
        Used to toggle the 'streaming' variable in the remote "run" process 
        '''
        self.stream.set()


class MultiChanDataSource(mp.Process):
    '''
    Multi-channel version of 'DataSource'
    '''
    def __init__(self, source, bufferlen=5, name=None, send_data_to_sink_manager=False, **kwargs):
        '''
        Parameters
        ----------
        source: class 
            lower-level class for interacting directly with the incoming data (e.g., plexnet)
        bufferlen: int
            Constrains the maximum amount of data history stored by the source
        name: string, optional, default=None
            Name of the sink, i.e., HDF table. If one is not provided, it will be inferred based
            on the name of the source module
        send_data_to_sink_manager: boolean, optional, default=False
            Flag to indicate whether data should be saved to a sink (e.g., HDF file)            
        kwargs: dict, optional, default = {}
            For the multi-channel data source, you MUST specify a 'channels' keyword argument
            Note that kwargs['channels'] does not need to a list of integers,
            it can also be a list of strings.
        '''

        super(MultiChanDataSource, self).__init__()
        if name is not None:
            self.name = name
        else:
            self.name = source.__module__.split('.')[-1]
        self.filter = None
        self.source = source
        self.source_kwargs = kwargs
        self.bufferlen = bufferlen
        self.max_len = int(bufferlen * self.source.update_freq)
        self.channels = kwargs['channels']
        self.chan_to_row = dict()
        for row, chan in enumerate(self.channels):
            self.chan_to_row[chan] = row
        
        self.n_chan = len(self.channels)
        dtype = self.source.dtype  # e.g., np.dtype('float') for LFP
        self.slice_size = dtype.itemsize
        self.idxs = shm.RawArray('l', self.n_chan)
        self.last_read_idxs = np.zeros(self.n_chan, dtype='int')
        rawarray = shm.RawArray('c', self.n_chan * self.max_len * self.slice_size)

        self.data = np.frombuffer(rawarray, dtype).reshape((self.n_chan, self.max_len))

        #self.fo2 = open('/storage/rawdata/test_rda_get.txt','w')
        #self.fo3 = open('/storage/rawdata/test_rda_run.txt','w')

        self.lock = mp.Lock()
        self.pipe, self._pipe = mp.Pipe()
        self.cmd_event = mp.Event()
        self.status = mp.Value('b', 1)
        self.stream = mp.Event()
        self.data_has_arrived = mp.Value('b', 0)

        self.methods = set(n for n in dir(source) if inspect.ismethod(getattr(source, n)))

        self.send_data_to_sink_manager = send_data_to_sink_manager
        if self.send_data_to_sink_manager:
            self.send_to_sinks_dtype = np.dtype([('chan'+str(chan), dtype) for chan in kwargs['channels']])
            self.next_send_idx = mp.Value('l', 0)
            self.wrap_flags = shm.RawArray('b', self.n_chan)  # zeros/Falses by default

    def register_supp_hdf(self):
        self.supp_hdf = supp_hdf.SupplementaryHDF(self.channels, self.send_to_sinks_dtype, self.source)
        
    def start(self, *args, **kwargs):
        '''
        From Python's docs on the multiprocessing module:
            Start the process's activity.
            This must be called at most once per process object. It arranges for the object's run() method to be invoked in a separate process.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        self.sinks = sink_manager = sink.SinkManager.get_instance()
        super(MultiChanDataSource, self).start(*args, **kwargs)

    def run(self):
        '''
        Main function executed by the mp.Process object. This function runs in the *remote* process, not in the main process
        '''
        print(("Starting datasource %r" % self.source))
        if self.send_data_to_sink_manager:
            print(("Registering Supplementary HDF file for datasource %r" % self.source))
            self.register_supp_hdf()

        try:
            system = self.source(**self.source_kwargs)
            system.start()

        except Exception as e:
            print(e)
            self.status.value = -1

        streaming = True
        size = self.slice_size
        while self.status.value > 0:
            if self.cmd_event.is_set():
                cmd, args, kwargs = self._pipe.recv()
                self.lock.acquire()
                try:
                    if cmd == "getattr":
                        ret = getattr(system, args[0])
                    else:
                        ret = getattr(system, cmd)(*args, **kwargs)
                except Exception as e:
                    traceback.print_exc()
                    ret = e
                self.lock.release()
                self._pipe.send(ret)
                self.cmd_event.clear()

            if self.stream.is_set():
                self.stream.clear()
                streaming = not streaming
                if streaming:
                    self.idx.value = 0
                    system.start()
                else:
                    system.stop()

            if streaming:
                # system.get() must return a tuple (chan, data), where: 
                #   chan is the the channel number
                #   data is a numpy array with a dtype (or subdtype) of
                #   self.source.dtype
                #print 'before get'
                

                chan, data = system.get()
                #self.fo3.write(str(data[0][0]) + ' ' + str(time.time()) + ' \n')

                #print 'after get'
                if data is not None:
                    try:
                        self.lock.acquire()
                        
                        try:
                            row = self.chan_to_row[chan]  # row in ringbuffer corresponding to this channel
                        except KeyError:
                            # print 'data source was not configured to get data on channel', chan
                            pass
                        else:
                            n_pts = len(data)
                            max_len = self.max_len

                            if n_pts > max_len:
                                data = data[-max_len:]
                                n_pts = max_len

                            idx = self.idxs[row] # for this channel, idx in ringbuffer
                            if idx + n_pts <= self.max_len:
                                self.data[row, idx:idx+n_pts] = data
                                idx = (idx + n_pts)
                                if idx == self.max_len:
                                    idx = 0
                                    if self.send_data_to_sink_manager:
                                        self.wrap_flags[row] = True
                            else: # need to write data at both end and start of buffer
                                self.data[row, idx:] = data[:max_len-idx]
                                self.data[row, :n_pts-(max_len-idx)] = data[max_len-idx:]
                                idx = n_pts-(max_len-idx)
                                if self.send_data_to_sink_manager:
                                    self.wrap_flags[row] = True
                            self.idxs[row] = idx

                        self.lock.release()

                        # Set the flag indicating that data has arrived from the source
                        self.data_has_arrived.value = 1
                    except Exception as e:
                        print(e)

                    if self.send_data_to_sink_manager:
                        self.lock.acquire()

                        # check if there is at least one column of data that
                        # has not yet been sent to the sink manager
                        if all(self.next_send_idx.value < idx + int(flag)*self.max_len for (idx, flag) in zip(self.idxs, self.wrap_flags)):
                            start_idx = self.next_send_idx.value
                            if not all(self.wrap_flags):

                                # look at minimum value of self.idxs only 
                                # among channels which have not wrapped, 
                                # in order to determine end_idx
                                end_idx = np.min([idx for (idx, flag) in zip(self.idxs, self.wrap_flags) if not flag])
                                idxs_to_send = list(range(start_idx, end_idx))
                            else:
                                min_idx = np.min(self.idxs[:])
                                idxs_to_send = list(range(start_idx, self.max_len)) + list(range(0, min_idx))
                                
                                for row in range(self.n_chan):
                                    self.wrap_flags[row] = False

                            # Old way to send data to the sink manager, one column at a time
                            # for idx in idxs_to_send:
                            #     data = np.array([tuple(self.data[:, idx])], dtype=self.send_to_sinks_dtype)
                            #     print "data shape"
                            #     print data.shape
                            #     self.sinks.send(self.name, data)

                            # # # New way to send data (in blocks) (update 1/12/2016): all columns at a time
                            #ix_ = np.ix_(np.arange(self.data.shape[0]), idxs_to_send)
                            #data = np.array(self.data[ix_], dtype=self.send_to_sinks_dtype)
                            #self.sinks.send(self.name, data)

                            #Newest way to send data to the supp hdf file, all columns at a time (1/21/2016)
                            data = np.array(list(map(tuple, self.data[:, idxs_to_send].T)), dtype = self.send_to_sinks_dtype)
                            self.supp_hdf.add_data(data)


                            self.next_send_idx.value = np.mod(idxs_to_send[-1] + 1, self.max_len)

                        self.lock.release()
            else:
                time.sleep(.001)
        
        if hasattr(self, "supp_hdf"):
            self.supp_hdf.close_data()
            print('end of supp hdf')

        system.stop()
        print(("ended datasource %r" % self.source))



    def get(self, n_pts, channels, **kwargs):
        '''
        Return the most recent n_pts of data from the specified channels.

        Parameters
        ----------
        n_pts : int
            Number of data points to read
        channels : iterable
            Channels from which to read

        Returns
        -------
        list of np.recarray objects
            Datatype of each record array is the dtype of the DataSourceSystem
        '''
        if self.status.value <= 0:
            raise Exception('\n\nError starting datasource: %s\n\n' % self.name)

        self.lock.acquire()   
   
        # these channels must be a subset of the channels passed into __init__
        n_chan = len(channels)
        data = np.zeros((n_chan, n_pts), dtype=self.source.dtype)

        if n_pts > self.max_len:
            n_pts = self.max_len
        # print "channels", channels[-1]
        for chan_num, chan in enumerate(channels):
            try:
                row = self.chan_to_row[chan]
            except KeyError:
                print(('data source was not configured to get data on channel', chan))
            else:  # executed if try clause does not raise a KeyError
                idx = self.idxs[row]
                if idx >= n_pts:  # no wrap-around required
                    data[chan_num, :] = self.data[row, idx-n_pts:idx]
                else:
                    data[chan_num, :n_pts-idx] = self.data[row, -(n_pts-idx):]
                    data[chan_num, n_pts-idx:] = self.data[row, :idx]
                self.last_read_idxs[row] = idx
        #print (data['data'])
        #self.fo2.write(str(data[0,0]['data']) + ' ' + str(len(data[0,:]['data'])) + ' ' + str(time.time()) + ' \n')
        self.lock.release()

        if self.filter is not None:
            return self.filter(data, **kwargs)
        return data

    def get_new(self, channels, **kwargs):
        '''
        Return the new (unread) data from the specified channels.

        Parameters
        ----------
        channels : iterable
            Channels from which to read        
        kwargs : optional kwargs 
            To be passed to self.filter, if it is listed

        Returns
        -------
        list of np.recarray objects
            Datatype of each record array is the dtype of the DataSourceSystem        
        '''
        if self.status.value <= 0:
            raise Exception('\n\nError starting datasource: %s\n\n' % self.name)

        self.lock.acquire()
        
        # these channels must be a subset of the channels passed into __init__
        n_chan = len(channels)
        data = []

        for chan in channels:
            try:
                row = self.chan_to_row[chan]
            except KeyError:
                print(('data source was not configured to get data on channel', chan))
                data.append(None)
            else:  # executed if try clause does not raise a KeyError
                idx = self.idxs[row]
                last_read_idx = self.last_read_idxs[row]
                if last_read_idx <= idx:  # no wrap-around required
                    data.append(self.data[row, last_read_idx:idx])
                else:
                    data.append(np.hstack((self.data[row, last_read_idx:], self.data[row, :idx])))
            self.last_read_idxs[row] = idx

        self.lock.release()

        if self.filter is not None:
            return self.filter(data, **kwargs)
        return data

    def pause(self):
        '''
        Used to toggle the 'streaming' variable in the remote "run" process 
        '''
        self.stream.set()

    def check_if_data_has_arrived(self):
        '''
        '''
        return self.data_has_arrived.value

    def stop(self):
        '''
        Set self.status.value to negative so that the while loop in self.run() terminates
        '''
        self.status.value = -1
        # self.fo2.close()
        # self.fo3.close()
    
    def __del__(self):
        '''
        Make sure the remote process stops if the Source object is destroyed
        '''
        self.stop()

    def __getattr__(self, attr):
        '''
        Try to retreive attributes from the remote DataSourceSystem if the are not found in the proximal Source object

        Parameters
        ----------
        attr : string 
            Name of attribute to retreive

        Returns
        -------
        object
            The arbitrary value associated with the named attribute, if it exists.
        '''
        if attr == 'methods':
            return super().__getattr__(attr)
        if attr in self.methods:
            return FuncProxy(attr, self.pipe, self.cmd_event)
        elif not attr.beginsWith("__"):
            print(("getting attribute %s" % attr))
            self.pipe.send(("getattr", (attr,), {}))
            self.cmd_event.set()
            return self.pipe.recv()
        raise AttributeError(attr)




import zmq
from riglib import experiment
from features.eyetracker_features import PupilLabStreaming
from riglib.pupillabs import System, NoSurfaceTracking
from riglib.pupillabs.pupillab_timesync import setup_pupil_remote_connection, request_pupil_time
from datetime import datetime
import time
import numpy as np
import os
import natnet

import unittest


class TestPupillabs(unittest.TestCase):

    # def test_client(self):
    #     socket = setup_pupil_remote_connection(ip_adress='128.95.215.191')
    #     time = request_pupil_time(socket)
    #     print(time)

    @unittest.skip("")
    def test_datasource(self):
        eyedata = NoSurfaceTracking()
        eyedata.start()
        time.sleep(0.5)

        data = eyedata.get()
        print(data)

        time.sleep(0.5)

        data = eyedata.get()
        print(data)

        eyedata.stop()

    def test_datasource_system(self):
        from riglib import source
        motiondata = source.DataSource(NoSurfaceTracking)
        motiondata.start()
        time.sleep(2)
        
        # Count packet rate
        count = 0
        duration = 5
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < duration:
            data = motiondata.get()
            if len(data) > 0:
                count += 1
            time.sleep(0.001)

        print(f"Packet rate: {count/duration} Hz")

        motiondata.stop()

        print(data)

if __name__ == '__main__':
    unittest.main()


import time
import numpy as np
import tempfile
import glob
np.set_printoptions(suppress=True)
from riglib.sink import sinks, PrintSink
from riglib import source
from riglib import motiontracker, hdfwriter, nidaq

Motion = motiontracker.make(38, motiontracker.System)
datasource = source.DataSource(Motion)
tf = tempfile.NamedTemporaryFile()
sinks.start(hdfwriter.HDFWriter, filename=tf.name)
sinks.register(Motion)


datasource.start()
time.sleep(10)
datasource.stop()
sinks.stop()
time.sleep(2)

import tables
h5 = tables.openFile(tf.name)
caldat = h5.root.motiontracker[1024:, -6:,:3]

d = glob.glob("/storage/calibrations/alignment*.npz")
motiontracker.make_autoalign_reference(caldat, "/storage/calibrations/alignment%d.npz"%len(d))



from .comms import QuattroOtlight
import numpy as np
from riglib.source import DataSourceSystem

'''
quattrocento streaming sources
'''

def multi_chan_generator(data_block, downsample=1):
    for idx in range(data_block.shape[1]):
        yield (idx+1, data_block[::downsample, idx]) # yield one channel at a time

class EMG(DataSourceSystem):
    '''
    Wrapper class for QuattroOtlight compatible with using in DataSource for
    buffering neural data. Compatible with riglib.source.MultiChanDataSource
    '''
    update_freq = 2048.
    dtype = np.dtype('int16')
    emg_refresh_freq = 32
    n_arrays = 1

    def __init__(self, *args, **kwargs):

        # Configure channels based on how many 64-channel arrays are connected to the MULTI IN ports
        channels = [16 * 8, 16 * 8 + 64 * self.n_arrays]
        self.qt = QuattroOtlight(host='128.95.215.191', refresh_freq=self.emg_refresh_freq, 
                                 emg_ch_range=channels, sample_freq=self.update_freq)

    def start(self):
        self.qt.setup()
        
        # Start with an empty generator
        self.gen = iter(())

    def stop(self):
        self.qt.tear_down()
        
    def get(self):
        '''
        Retrieve a packet from the host and split it into individual channels
        '''
        try:
            return next(self.gen)
        except StopIteration:
            emg, aux, samples = self.qt.read_emg()
            data = np.hstack([emg, aux, samples])
            self.gen = multi_chan_generator(data)
            return next(self.gen)


"""
Robot kinematic operations.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
from . import jacobian as Jac
from numpy.linalg import norm
from numpy.linalg import pinv
from math import *


def fkine(robot, q):
    """
    Computes the forward kinematics for each joint space point defined by C{q}.
    ROBOT is a robot object.

    For an n-axis manipulator C{q} is an n element vector or an m x n matrix of
    robot joint coordinates.

    If C{q} is a vector it is interpretted as the generalized joint coordinates, and
    C{fkine} returns a 4x4 homogeneous transformation for the tool of
    the manipulator.

    If C{q} is a matrix, the rows are interpretted as the generalized 
    joint coordinates for a sequence of points along a trajectory.  q[i,j] is
    the j'th joint parameter for the i'th trajectory point.  In this case
    C{fkine} returns a list of matrices for each point
    along the path.

    The robot's base or tool transform, if present, are incorporated into the
    result.
    
    @type robot: Robot instance
    @param robot: The robot
    @type q: vector
    @param q: joint coordinate
    @see: L{Link}, L{Robot}, L{ikine}
    """

    q = mat(q)
    n = robot.n
    if numrows(q)==1 and numcols(q)==n:
        t = robot.base
        for i in range(0,n):
            t = t * robot.links[i].tr(q[0,i])
        t = t * robot.tool
        return t
    else:
        if numcols(q) != n:
            raise Exception('bad data')
        t = []
        for qv in q:        # for each trajectory point
            tt = robot.base
            for i in range(0,n):
                tt = tt * robot.links[i].tr(qv[0,i])
            t.append(tt*robot.tool)
        return t



def ikine(robot, tr, q=None, m=None):
    """
    Inverse manipulator kinematics.
    Computes the joint coordinates corresponding to the end-effector transform C{tr}.
    Typically invoked as

        - Q = IKINE(ROBOT, T)
        - Q = IKINE(ROBOT, T, Q)
        - Q = IKINE(ROBOT, T, Q, M)

    Uniqueness
    ==========
    Note that the inverse kinematic solution is generally not unique, and 
    depends on the initial guess C{q} (which defaults to 0).

    Iterative solution
    ==================
    Solution is computed iteratively using the pseudo-inverse of the
    manipulator Jacobian.

    Such a solution is completely general, though much less efficient 
    than specific inverse kinematic solutions derived symbolically.

    This approach allows a solution to obtained at a singularity, but 
    the joint angles within the null space are arbitrarily assigned.

    Operation on a trajectory
    =========================
    If C{tr} is a list of transforms (a trajectory) then the solution is calculated
    for each transform in turn.  The return values is a matrix with one row for each
    input transform.  The initial estimate for the iterative solution at 
    each time step is taken as the solution from the previous time step.

    Fewer than 6DOF
    ===============
    If the manipulator has fewer than 6 DOF then this method of solution
    will fail, since the solution space has more dimensions than can
    be spanned by the manipulator joint coordinates.  In such a case
    it is necessary to provide a mask matrix, C{m}, which specifies the 
    Cartesian DOF (in the wrist coordinate frame) that will be ignored
    in reaching a solution.  The mask matrix has six elements that
    correspond to translation in X, Y and Z, and rotation about X, Y and
    Z respectively.  The value should be 0 (for ignore) or 1.  The number
    of non-zero elements should equal the number of manipulator DOF.

    For instance with a typical 5 DOF manipulator one would ignore
    rotation about the wrist axis, that is, M = [1 1 1 1 1 0].


    @type robot: Robot instance
    @param robot: The robot
    @type tr: homgeneous transformation
    @param tr: End-effector pose
    @type q: vector
    @param q: initial estimate of joint coordinate
    @type m: vector
    @param m: mask vector
    @rtype: vector
    @return: joint coordinate
    @see: L{fkine}, L{tr2diff}, L{jacbo0}, L{ikine560}
    """
     
    #solution control parameters
    
    ilimit = 1000
    stol = 1e-12
    n = robot.n
    
    if q == None:
        q = mat(zeros((n,1)))
    else:
        q = mat(q).flatten().T
        
    if q != None and m != None:
        m = mat(m).flatten().T
        if len(m)!=6:
            error('Mask matrix should have 6 elements')
        if len(m.nonzero()[0].T)!=robot.n:
            error('Mask matrix must have same number of 1s as robot DOF')
    else:
        if n<6:
            print('For a manipulator with fewer than 6DOF a mask matrix argument should be specified')
        m = mat(ones((6,1)))

    if isinstance(tr, list):
        #trajectory case
        qt = mat(zeros((0,n)))
        for T in tr:
            nm = 1
            count = 0
            while nm > stol:
                e = multiply(tr2diff( fkine(robot, q.T), T), m)
                dq = pinv(Jac.jacob0(robot,q.T))*e
                q += dq
                nm = norm(dq)
                count += 1
                if count > ilimit:
                    print('i=',i,'   nm=',nm)
                    error("Solution wouldn't converge")
            qt = vstack( (qt, q.T) )
        return qt;
    elif ishomog(tr):
        #single xform case
        nm = 1
        count = 0
        while nm > stol:
            e = multiply( tr2diff(fkine(robot,q.T),tr), m )
            dq = pinv(Jac.jacob0(robot, q.T)) * e
            q += dq;
            nm = norm(dq)
            count += 1
            if count > ilimit:
                error("Solution wouldn't converge")
        qt = q.T
        return qt
    else:
        error('tr must be 4*4 matrix')


def ikine560(robot, T, configuration=''):
    """
    Inverse kinematics for Puma 560-like robot, ie. 6-axis with a spherical wrist.

    The optional C{configuration} argument specifies the configuration of the arm in
    the form of a string containing one or more of the configuration codes
       - 'l' or 'r'    lefty/righty
       - 'u' or 'd'    elbow
       - 'n' or 'f'    wrist flip or noflip.

    The default configuration is 'lun'.
    
    Reference
    =========

    Inverse kinematics for a PUMA 560 based on the equations by Paul and Zhang
    From The International Journal of Robotics Research
    Vol. 5, No. 2, Summer 1986, p. 32-44.

    @author: Robert Biro (gt2231a@prism.gatech.edu) with Gary Von McMurray, GTRI/ATRP/IIMB, Georgia Institute of Technology, 2/13/95.

    @type robot: Robot instance
    @param robot: The robot
    @type T: homgeneous transformation
    @param T: End-effector pose
    @type configuration: string
    @param configuration: manipulator configuration comprising the letters: lrudnf
    @rtype: vector
    @return: joint coordinate
    """

    if robot.n != 6:
        error('Solution only applicable for 6DOF manipulator');

    if robot.mdh:
        error('Solution only applicable for standard DH conventions');


    # recurse over a list of transforms
    if isinstance(T, list):
        theta = [];
        for t in T:
            theta.append( ikine560(robot, t, configuration) );

        return theta;

    if not ishomog(T):
        error('T is not a homog xform');
        
    L = robot.links;
    a1 = L[0].A;
    a2 = L[1].A;
    a3 = L[2].A;

    for i in range(3, 6):
        if L[i].A != 0:
            error('wrist is not spherical')

    d1 = L[0].D;
    d2 = L[1].D;
    d3 = L[2].D;
    d4 = L[3].D;


    # undo base transformation
    T = linalg.inv(robot.base) * T;

    # The following parameters are extracted from the Homogeneous 
    # Transformation as defined in equation 1, p. 34

    Ox = T[0,1];
    Oy = T[1,1];
    Oz = T[2,1];

    Ax = T[0,2];
    Ay = T[1,2];
    Az = T[2,2];

    Px = T[0,3];
    Py = T[1,3];
    Pz = T[2,3];

    # The configuration parameter determines what n1,n2,n4 values are used
    # and how many solutions are determined which have values of -1 or +1.

    configuration = configuration.lower();

    n1 = -1;    # L
    n2 = -1;    # U
    n4 = -1;    # N
    if 'l' in configuration:
        n1 = -1;

    if 'r' in configuration:
        n1 = 1;

    if 'u' in configuration:
        if n1 == 1:
            n2 = 1;
        else:
            n2 = -1;

    if 'd' in configuration:
        if n1 == 1:
            n2 = -1;
        else:
            n2 = 1;

    if 'n' in configuration:
        n4 = 1;

    if 'f' in configuration:
        n4 = -1;


    theta = zeros( (6,1) );
    
    #
    # Solve for theta(1)
    # 
    # r is defined in equation 38, p. 39.
    # theta(1) uses equations 40 and 41, p.39, 
    # based on the configuration parameter n1
    #

    r = sqrt(Px**2 + Py**2);
    if n1 == 1:
        theta[0] = atan2(Py,Px) + asin(d3/r);
    else:
        theta[0] = atan2(Py,Px) + pi - asin(d3/r);


    #
    # Solve for theta(2)
    #
    # V114 is defined in equation 43, p.39.
    # r is defined in equation 47, p.39.
    # Psi is defined in equation 49, p.40.
    # theta(2) uses equations 50 and 51, p.40, based on the configuration 
    # parameter n2
    #

    V114 = Px*cos(theta[0]) + Py*sin(theta[0]);
    r = sqrt(V114**2 + Pz**2);

    x = (a2**2-d4**2-a3**2+V114**2+Pz**2) / (2.0*a2*r);
    if abs(x) > 1:
        error('point not reachable');
    Psi = acos(x);

    theta[1] = atan2(Pz,V114) + n2*Psi;

    #
    # Solve for theta(3)
    #
    # theta(3) uses equation 57, p. 40.
    #

    num = cos(theta[1])*V114+sin(theta[1])*Pz-a2;
    den = cos(theta[1])*Pz - sin(theta[1])*V114;
    theta[2] = atan2(a3,d4) - atan2(num, den);

    #
    # Solve for theta(4)
    #
    # V113 is defined in equation 62, p. 41.
    # V323 is defined in equation 62, p. 41.
    # V313 is defined in equation 62, p. 41.
    # theta(4) uses equation 61, p.40, based on the configuration 
    # parameter n4
    #

    V113 = cos(theta[0])*Ax + sin(theta[0])*Ay;
    V323 = cos(theta[0])*Ay - sin(theta[0])*Ax;
    V313 = cos(theta[1]+theta[2])*V113 + sin(theta[1]+theta[2])*Az;
    theta[3] = atan2((n4*V323),(n4*V313));
    #[(n4*V323),(n4*V313)]

    #
    # Solve for theta(5)
    #
    # num is defined in equation 65, p. 41.
    # den is defined in equation 65, p. 41.
    # theta(5) uses equation 66, p. 41.
    #
     
    num = -cos(theta[3])*V313 - V323*sin(theta[3]);
    den = -V113*sin(theta[1]+theta[2]) + Az*cos(theta[1]+theta[2]);
    theta[4] = atan2(num,den);
    #[num den]

    #
    # Solve for theta(6)
    #
    # V112 is defined in equation 69, p. 41.
    # V122 is defined in equation 69, p. 41.
    # V312 is defined in equation 69, p. 41.
    # V332 is defined in equation 69, p. 41.
    # V412 is defined in equation 69, p. 41.
    # V432 is defined in equation 69, p. 41.
    # num is defined in equation 68, p. 41.
    # den is defined in equation 68, p. 41.
    # theta(6) uses equation 70, p. 41.
    #

    V112 = cos(theta[0])*Ox + sin(theta[0])*Oy;
    V132 = sin(theta[0])*Ox - cos(theta[0])*Oy;
    V312 = V112*cos(theta[1]+theta[2]) + Oz*sin(theta[1]+theta[2]);
    V332 = -V112*sin(theta[1]+theta[2]) + Oz*cos(theta[1]+theta[2]);
    V412 = V312*cos(theta[3]) - V132*sin(theta[3]);
    V432 = V312*sin(theta[3]) + V132*cos(theta[3]);
    num = -V412*cos(theta[4]) - V332*sin(theta[4]);
    den = - V432;
    theta[5] = atan2(num,den);
    #[num den]
    
    return mat(theta).T;



'''
Code for the triggering stimulation from the Master-8 voltage-programmable stimulator
'''
import comedi

###### CONSTANTS
sec_per_min = 60

"""
Define one class to send TTL pulse for one cycle of stimulation (StimulationPulse).  Define second class to implement this
at a fixed rate.

Left off at line 79.  Need to figure out how we re-trigger starting the stimulus pulse train again.
"""



class TTLStimulation(object):
    '''During the stimulation phase, send a timed TTL pulse to the Master-8 stimulator'''
    hold_time = float(1)
    stimulation_pulse_length = float(0.2*1e6)
    stimulation_frequency = float(3)
    
    status = dict(
        pulse = dict(pulse_end="interpulse_period", stop=None),
        interpulse_period = dict(another_pulse="pulse", pulse_train_end="pulse_off", stop=None),
        pulse_off= dict(stop=None)
    )

    com = comedi.comedi_open('/dev/comedi0')
    pulse_count = 0  #initializing number of pulses that have occured

    def __init__(self, *args, **kwargs):
        super(TTLStimulation, self).__init__(*args, **kwargs)
        number_of_pulses = int(self.hold_time*self.stimulation_frequency)   # total pulses during a stimulation pulse train, assumes hold_time is in s

    def init(self):
        super(TTLStimulation, self).init()

    #### TEST FUNCTIONS ####

    def _test_pulse_end(self, ts):
        #return true if time has been longer than the specified pulse duration
        pulse_length = self.stimulation_pulse_length*1e-6    # assumes stimulation_pulse_length is in us
        return ts>=pulse_length

    def _test_another_pulse(self,ts):
        interpulse_time = (1/self.stimulationfrequency) - self.stimulation_pulse_length*1e-6    # period minus the duration of a pulse
        return ts>=interpulse_time

    def _test_pulse_train_end(self,ts):
        return (self.pulse_count > number_of_pulses)     # end train if number of pulses is completed or if the animal ends holding early


    #### STATE FUNCTIONS ####

    def _start_pulse(self):
        '''
        At the start of the stimulation state, send TTL pulse
        '''

        #super(TTLStimulation, self)._start_pulse()
        subdevice = 0
        write_mask = 0x800000
        val = 0x800000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)
        self.pulse_count = self.pulse_count + 1
        #self.stimulation_start = self.get_time() - self.start_time

    def _end_pulse(self):
        subdevice = 0
        write_mask = 0x800000
        val = 0x000000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)

    def _start_interpulse_period(self):
        super(TTLStimulation, self)._start_interpulse_period()
        subdevice = 0
        write_mask = 0x800000
        val = 0x000000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)

    def _end_interpulse_period(self):
        super(TTLStimulation, self)._end_interpulse_period()
        subdevice = 0
        write_mask = 0x800000
        val = 0x000000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)




#!/usr/bin/python
'''
Code for interacting with the positioner microcontroller
'''
import serial
import time
import struct
import numpy as np
import re
from riglib.experiment import Experiment, Sequence, FSMTable, StateTransitions
import random

import socket
import select

dir_lut = dict(x={0:0, -1:0, 1:1}, 
    y={0:0, -1:0, 1:1}, 
    z={0:1, -1:1, 1:0}, # convention flipped for z-stage
) 

class Positioner(object):
    def __init__(self, dev='/dev/arduino_positioner'):
        self.port = serial.Serial(dev, baudrate=115200)
        self.port.flushInput()

    def _parse_resp(self, resp):
        resp = resp.rstrip()
        limits = list(map(int, resp[-6:]))
        return limits

    def poll_limit_switches(self, N=100):
        while 1:
            time.sleep(0.1)
            self.port.write('\n')
            raw_resp = self.port.readline()
            print("limit switches", self._parse_resp(raw_resp))

    def read_limit_switches(self):
        self.port.write('\n')
        raw_resp = self.port.readline()
        return self._parse_resp(raw_resp)

    def wake_motors(self):
        self.port.write('w\n')
        # self.port.readline()

    def sleep_motors(self):
        print("sleep motors")
        self.port.write('s\n')
        # self.port.readline()

    def step_motors(self, step_x, step_y, step_z, dir_x, dir_y, dir_z):
        cmd_data = 0
        cmd_step_data = step_x | (step_y << 1) | (step_z << 2)
        cmd_dir_data = dir_x | (dir_y << 1) | (dir_z << 2)
        cmd_data = cmd_step_data | (cmd_dir_data << 4)
        cmd = 'm' + struct.pack('B', cmd_data) + '\n'
        #print cmd_data, cmd
        self.port.write(cmd)

    def move(self, n_steps_x, n_steps_y, n_steps_z):
        self.wake_motors()
        limits = self._parse_resp(self.port.readline())
        time.sleep(1)

        dir_x = dir_lut['x'][np.sign(n_steps_x)]
        dir_y = dir_lut['y'][np.sign(n_steps_y)]
        dir_z = dir_lut['z'][np.sign(n_steps_z)]

        n_steps_sent_x = 0
        n_steps_sent_y = 0
        n_steps_sent_z = 0

        k = 0
        while (abs(n_steps_x) > n_steps_sent_x) or (abs(n_steps_y) > n_steps_sent_y) or (abs(n_steps_z) > n_steps_sent_z):
            if k % 10 == 0: print(k)
            step_x = int(n_steps_sent_x < abs(n_steps_x))
            step_y = int(n_steps_sent_y < abs(n_steps_y))
            step_z = int(n_steps_sent_z < abs(n_steps_z))
            #print step_x, step_y, step_z, dir_x, dir_y, dir_z
            self.step_motors(step_x, step_y, step_z, dir_x, dir_y, dir_z)
            limits = self._parse_resp(self.port.readline())
            k += 1

            n_steps_sent_x += step_x
            n_steps_sent_y += step_y
            n_steps_sent_z += step_z
        
        self.sleep_motors()

    def old_move(self, n_steps_x, n_steps_y, n_steps_z):
        self.wake_motors()
        try:
            time.sleep(1)

            dir_x = dir_lut['x'][np.sign(n_steps_x)]
            dir_y = dir_lut['y'][np.sign(n_steps_y)]
            dir_z = dir_lut['z'][np.sign(n_steps_z)]

            n_steps_sent_x = 0
            n_steps_sent_y = 0
            n_steps_sent_z = 0

            k = 0
            while (abs(n_steps_x) > n_steps_sent_x) or (abs(n_steps_y) > n_steps_sent_y) or (abs(n_steps_z) > n_steps_sent_z):
                if k % 10 == 0: print(k)
                step_x = int(n_steps_sent_x < abs(n_steps_x))
                step_y = int(n_steps_sent_y < abs(n_steps_y))
                step_z = int(n_steps_sent_z < abs(n_steps_z))
                #print step_x, step_y, step_z, dir_x, dir_y, dir_z
                self.step_motors(step_x, step_y, step_z, dir_x, dir_y, dir_z)
                limits = self._parse_resp(self.port.readline())
                k += 1

                n_steps_sent_x += step_x
                n_steps_sent_y += step_y
                n_steps_sent_z += step_z
            
        except:
            import traceback
            traceback.print_exc()
        finally:
            self.sleep_motors()

    def move2(self):
        self.wake_motors()
        for k in range(200):
            self.step_motors(1,0,0,0,0,0)
            limits = self._parse_resp(self.port.readline())
        self.sleep_motors()

    def go_to_min(self, verbose=False):
        can_move = self.read_limit_switches()
        x_can_decrease = can_move[0]
        y_can_decrease = can_move[2]
        z_can_decrease = can_move[4]

        dir_x = dir_lut['x'][-1]
        dir_y = dir_lut['y'][-1]
        dir_z = dir_lut['z'][-1]

        n_steps_sent_x = 0
        n_steps_sent_y = 0
        n_steps_sent_z = 0
        if x_can_decrease or y_can_decrease or z_can_decrease:
            self.wake_motors()

        try:
            k = 0
            while x_can_decrease or y_can_decrease or z_can_decrease:
                step_x = int(x_can_decrease)
                step_y = int(y_can_decrease)
                step_z = int(z_can_decrease)
                if verbose:
                    print(step_x, step_y, step_z)

                self.step_motors(step_x, step_y, step_z, dir_x, dir_y, dir_z)
                can_move = self._parse_resp(self.port.readline())

                x_can_decrease = can_move[0]
                y_can_decrease = can_move[2]
                z_can_decrease = can_move[4]

                n_steps_sent_x += step_x
                n_steps_sent_y += step_y
                n_steps_sent_z += step_z
                k += 1
        finally:
            self.sleep_motors()

        return n_steps_sent_x, n_steps_sent_y, n_steps_sent_z

    def go_to_max(self, verbose=False):
        can_move = self.read_limit_switches()
        x_can_increase = can_move[1]
        y_can_increase = can_move[3]
        z_can_increase = can_move[5]

        dir_x = dir_lut['x'][1]
        dir_y = dir_lut['y'][1]
        dir_z = dir_lut['z'][1]

        n_steps_sent_x = 0
        n_steps_sent_y = 0
        n_steps_sent_z = 0
        if x_can_increase or y_can_increase or z_can_increase:
            self.wake_motors()

        try:
            k = 0
            while x_can_increase or y_can_increase or z_can_increase:
                step_x = int(x_can_increase)
                step_y = int(y_can_increase)
                step_z = int(z_can_increase)
                if verbose:
                    print(step_x, step_y, step_z)

                self.step_motors(step_x, step_y, step_z, dir_x, dir_y, dir_z)
                can_move = self._parse_resp(self.port.readline())

                x_can_increase = can_move[1]
                y_can_increase = can_move[3]
                z_can_increase = can_move[5]

                n_steps_sent_x += step_x
                n_steps_sent_y += step_y
                n_steps_sent_z += step_z
                k += 1
        finally:
            self.sleep_motors()

        return n_steps_sent_x, n_steps_sent_y, n_steps_sent_z

    def continuous_move(self, n_steps_x, n_steps_y, n_steps_z):
        self.start_continuous_move(n_steps_x, n_steps_y, n_steps_z)
        return self.end_continuous_move()

    def start_continuous_move(self, n_steps_x, n_steps_y, n_steps_z):
        '''
        Same as 'continuous_move', but without blocking for a response/movement to finish before the function returns
        '''
        self.wake_motors()
        msg = 'c' + struct.pack('>hhh', n_steps_x, n_steps_y, n_steps_z) + '\n'
        self.port.write(msg)

        self.motor_dir = np.array([np.sign(n_steps_x), np.sign(n_steps_y), np.sign(n_steps_z)])

    def end_continuous_move(self, stiff=False):
        '''
        Cleanup part of 'continuous_move' after 'start_continuous_move' has been called
        '''
        movement_data = self.port.readline()

        try:
            m = re.match(".*?: (\d+), (\d+), (\d+)", movement_data)
            n_steps_actuated = list(map(int, [m.group(x) for x in [1,2,3]]))
        except:
            import traceback
            traceback.print_exc()
            print(movement_data)

        if not stiff:
            self.sleep_motors()

        return n_steps_actuated

    def calibrate(self, n_runs):
        '''
        Repeatedly go from min to max and back so the number of steps can be counted
        '''
        n_steps_min_to_max = [None]*n_runs
        n_steps_max_to_min = [None]*n_runs

        self.go_to_min()
        time.sleep(1)

        for k in range(n_runs):
            n_steps_min_to_max[k] = self.go_to_max()
            time.sleep(2)
            n_steps_max_to_min[k] = self.go_to_min()
            print("min to max")
            print(n_steps_min_to_max)
            print("max to min")
            print(n_steps_max_to_min)
            time.sleep(2)
            
        return n_steps_min_to_max, n_steps_max_to_min

    def data_available(self):
        return self.port.inWaiting()

# from features.generator_features import Autostart
class PositionerTaskController(Sequence):
    '''
    Interface between the positioner and the task interface. The positioner should run asynchronously
    so that the task event loop does not have to wait for a serial port response from the microcontroller.
    '''

    status = FSMTable(
        go_to_origin = StateTransitions(microcontroller_done='wait'),
        wait = StateTransitions(start_trial='move_target'),
        move_target = StateTransitions(microcontroller_done='reach', stoppable=False),
        reach = StateTransitions(time_expired='reward', new_target_set_remotely='move_target'),
        reward = StateTransitions(time_expired='wait'),
    )
    # status = dict(
    #     go_to_origin = dict(microcontroller_done='wait', stop=None),
    #     wait = dict(start_trial='move_target', stop=None),
    #     move_target = dict(microcontroller_done='reach'),
    #     reach = dict(time_expired='reward', stop=None),
    #     reward = dict(time_expired='wait'),
    # )


    state = 'go_to_origin'

    sequence_generators = ['random_target_calibration', 'xy_sweep']
    reward_time = 1
    reach_time = 1

    @staticmethod
    def random_target_calibration(n_blocks=10):
    #     # constants selected approximately from one subject's ROM
    #     targets = [
    #         (x_min, y_min, z_min),
    #         (x_max, y_min, z_min),
    #         (x_min, y_max, z_min),
    #         (x_min, y_min, z_max),
    #         (x_max, y_max, z_min),
    #         (x_max, y_min, z_max),
    #         (x_min, y_max, z_max),
    #         (x_max, y_max, z_max),
    #     ]

    #     trial_target_ls = []
    #     for k in range(n_blocks):
    #         random.shuffle(targets)
    #         for targ in targets:
    #             trial_target_ls.append(dict(int_target_pos=targ))

    #     # set the last target to be the origin since the purpose of this generator is to measure the drift in # of steps
    #     trial_target_ls.append(dict(int_target_pos=np.zeros(3)))
    #     return trial_target_ls

    # @staticmethod 
    # def calibration_targets(nblocks=1):
        targets = [
            (45, 34, 0),
            (50, 38, -25),
            (40, 35, 0),
            (40, 35, -25),            
            (30, 29, 0),
            (30, 29, -25),            
            (20, 35, 0),
            (20, 35, -25),
            # (10, 38, 0), # reachable?
            # (10, 38, -25), # reachable?
        ]
        trial_target_ls = []
        for k in range(n_blocks):
            random.shuffle(targets)
            for targ in targets:
                trial_target_ls.append(dict(int_target_pos=targ))

        # set the last target to be the origin since the purpose of this generator is to measure the drift in # of steps
        trial_target_ls.append(dict(int_target_pos=np.zeros(3)))
        return trial_target_ls        

    @staticmethod 
    def xy_sweep(z_min=-25, z_max=0, zpts=6):
        
        xy_target_locs = np.vstack([
            [8.20564516129, 37.6302083333],
            [9.61693548387, 34.1145833333],
            [15.1209677419, 31.1848958333],
            [15.5443548387, 34.5703125],
            [18.2258064516, 36.5234375],
            [23.4475806452, 34.7005208333],
            [22.8830645161, 32.3567708333],
            [23.0241935484, 29.1666666667],
            [28.9516129032, 34.8307291667],
            [28.9516129032, 32.2265625],
            [29.2338709677, 30.1432291667],
            [33.3266129032, 35.4166666667],
            [33.8911290323, 33.1380208333],
            [30.5040322581, 30.078125],
            [20.4838709677, 28.1901041667],
            [35.5846774194, 36.5885416667],
            [39.2540322581, 33.5286458333],
            [41.5120967742, 38.5416666667],
            [47.439516129,  37.6953125],
        ])

        trial_target_ls = []
        z_range = np.linspace(z_min, z_max, zpts)
        for zpt in z_range:
            for xy_targ in xy_target_locs:
                trial_target_ls.append(dict(int_target_pos=np.hstack([xy_targ, zpt])))

        return trial_target_ls

    def __init__(self, *args, **kwargs):
        '''
        Constructor for PositionerTaskController

        Parameters
        ----------
        # x_len : float
        #     measured distance the positioner can travel in the x-dimension
        # y_len : float
        #     measured distance the positioner can travel in the y-dimension
        # z_len : float
        #     measured distance the positioner can travel in the z-dimension
        dev : str, optional, default=/dev/ttyACM1
            Serial port to use to communicate with Arduino controller
        x_cm_per_rev : int, optional, default=12
            Number of cm traveled for one full revolution of the stepper motors in the x-dimension
        y_cm_per_rev : int, optional, default=12
            Number of cm traveled for one full revolution of the stepper motors in the y-dimension
        z_cm_per_rev : float, optional, default=7.6
            Number of cm traveled for one full revolution of the stepper motors in the z-dimension
        x_step_size : float, optional, default=0.25
            Microstepping mode in the x-dimension
        y_step_size : float, optional, default=0.25
            Microstepping mode in the y-dimension
        z_step_size : float, optional, default=0.25
            Microstepping mode in the z-dimension
    
        Returns
        -------
        PositionerTaskController instance
        '''

        # TODO make these input arguments
        positioner_dev = '/dev/arduino_positioner'

        # cm/rev based on measured data
        x_cm_per_rev = 12.4
        y_cm_per_rev = 12.4
        z_cm_per_rev = 8.0 

        x_step_size = 1./4
        y_step_size = 1./4
        z_step_size = 1./4

        self.loc = np.ones(3) * np.nan # position of the target relative to origin is unknown until the origin limit switches are hit
        self.pos_uctrl_iface = Positioner(dev=positioner_dev)
        # self.pos_uctrl_iface.sleep_motors()

        self.steps_from_origin = np.ones(3) * np.nan # cumulative number of steps taken from the origin. Unknown until the origin limit switches are hit.
        self.step_size = np.array([x_step_size, y_step_size, z_step_size], dtype=np.float64)
        self.cm_per_rev = np.array([x_cm_per_rev, y_cm_per_rev, z_cm_per_rev], dtype=np.float64)

        self.full_steps_per_rev = 200.

        super(PositionerTaskController, self).__init__(*args, **kwargs)

    def init(self):
        self.add_dtype('positioner_loc', np.float64, (3,))
        self.add_dtype('positioner_steps_from_origin', np.float64, (3,))
        super(PositionerTaskController, self).init()

        # open an rx_socket for reading new commands 
        import socket
        self.rx_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.rx_sock.bind(('localhost', 60005))

    def terminate(self):
        # close the rx socket used for reading remote commands
        super(PositionerTaskController, self).terminate()
        self.rx_sock.close()

    ##### Helper functions #####
    def _calc_steps_to_pos(self, target_pos):
        displ_cm = target_pos - self.loc

        # compute the number of steps needed to travel the desired displacement
        displ_rev = displ_cm / self.cm_per_rev
        displ_full_steps = displ_rev * self.full_steps_per_rev
        displ_microsteps = displ_full_steps / self.step_size 
        displ_microsteps = displ_microsteps.astype(int)
        return displ_microsteps

    def _steps_to_cm(self, n_steps_actuated):
        steps_moved = n_steps_actuated * self.step_size
        self.steps_from_origin += steps_moved
        return steps_moved / self.full_steps_per_rev * self.cm_per_rev        

    def _integrate_steps(self, n_steps_actuated, signs):
        steps_moved = n_steps_actuated * signs * self.step_size
        self.steps_from_origin += steps_moved
        self.loc += steps_moved / self.full_steps_per_rev * self.cm_per_rev

    def _test_microcontroller_done(self, *args, **kwargs):
        # check if any data has returned from the microcontroller interface
        # self.print_to_terminal("starting to check for data")
        bytes_avail = self.pos_uctrl_iface.data_available()

        # self.print_to_terminal(bytes_avail)

        # remember to actually read the data out of the buffer in an '_end' function
        return bytes_avail > 0

    # def update_report_stats(self, *args, **kwargs):
    #     super(PositionerTaskController, self).update_report_stats()
    #     self.reportstats['resp_bytes'] = self.pos_uctrl_iface.data_available()

    def _test_new_target_set_remotely(self, *args, **kwargs):
        # print "checking for new target set"
        socket_list = [self.rx_sock]
        # Get the list sockets which are readable
        read_sockets, write_sockets, error_sockets = select.select(socket_list , [], [], 0)
        if self.rx_sock in read_sockets:
            raw_data = self.rx_sock.recv(8*3)
            import struct
            new_pos = struct.unpack('ddd', raw_data)
            print("received new position!")
            print(new_pos)
            self._gen_int_target_pos = new_pos
            return True 
        else:
            return False

    ##### State transition functions #####
    def _start_go_to_origin(self):
        print("_start_go_to_origin")
        self.pos_uctrl_iface.start_continuous_move(1000, 1000, -1000)

    def _start_go_to_target(self,num_x,num_y,num_z):
        # AY modification - _start_go_to_origin sends the positioner to a predetermined location.  Need to be able to send it
        # different target locations for the different target positions.  Not currently working (also not implemented in tasklilst yet)
        print("_start_go_to_target")
        self.pos_uctrl_iface.start_continuous_move(num_x,num_y,num_z)

    def _start_reward(self):
        pass

    def _end_go_to_origin(self):
        steps_actuated = self.pos_uctrl_iface.end_continuous_move(stiff=True)

        self.loc = np.zeros(3)
        self.steps_from_origin = np.zeros(3)

    def _start_move_target(self):
        # calc number of steps from current pos to target pos
        displ_microsteps = self._calc_steps_to_pos(self._gen_int_target_pos)

        # send command to initiatem movement 
        self.pos_uctrl_iface.start_continuous_move(*displ_microsteps)

    def _end_move_target(self):
        # send command to kill motors
        steps_actuated = self.pos_uctrl_iface.end_continuous_move()
        self._integrate_steps(steps_actuated, self.pos_uctrl_iface.motor_dir)

    def _cycle(self):
        self.task_data['positioner_loc'] = self.loc
        self.task_data['positioner_steps_from_origin'] = self.steps_from_origin
        super(PositionerTaskController, self)._cycle()

    ### Old functions ###
    def go_to_origin(self):
        '''
        Tap the origin limit switches so the absolute position of the target can be estimated. 
        Run at initialization and/or periodically to correct for any accumulating errors.
        '''
        steps_moved = np.array(self.pos_uctrl_iface.continuous_move(-10000, -10000, 10000))
        step_signs = np.array([-1, -1, 1], dtype=np.float64) * self.step_size

        if not np.any(np.isnan(self.steps_from_origin)): # error accumulation correction
            self.steps_from_origin += step_signs * steps_moved

            # if no position errors were accumulated, then self.steps_from_origin should all be 0
            acc_error = self.steps_from_origin
            print("accumulated step errors")
            print(acc_error)

        self.loc = np.zeros(3)
        self.steps_from_origin = np.zeros(3)

    def go_to_position(self, target_pos):
        if np.any(np.isnan(self.loc)):
            raise Exception("System must be 'zeroed' before it can go to an absolute location!")

        displ_cm = target_pos - self.loc

        # compute the number of steps needed to travel the desired displacement
        displ_rev = displ_cm / self.cm_per_rev
        displ_full_steps = displ_rev * full_steps_per_rev
        displ_microsteps = displ_full_steps / self.step_size

        steps_moved = np.array(self.pos_uctrl_iface.continuous_move(*displ_microsteps))
        steps_moved = steps_moved * np.sign(displ_microsteps) * self.step_size

        self.steps_from_origin += steps_moved
        self.loc += steps_moved / self.full_steps_per_rev * self.cm_per_rev

    def run(self):
        '''
        Tell the positioner motors to turn off when the task ends
        '''
        try:
            super(PositionerTaskController, self).run()
        finally:
            self.pos_uctrl_iface.sleep_motors()
            

from .calib import *



import time
import os
import numpy as np
from riglib.experiment import traits
from riglib import quattrocento, source
from features.neural_sys_features import CorticalBMI,CorticalData
import traceback

class QuattBMI(CorticalBMI):
    '''
    BMI using quattrocento as the datasource.
    '''
    quatt_sampling_rate = traits.Float(2048, desc="Sampling rate of the emg data")
    n_emg_arrays = traits.Int(1, desc="Number of EMG arrays connected to the Quattrocento MULTI IN ports")

    def __init__(self, *args, **kwargs):
        '''
        This is a bit weird because normally only the readouts are streamed to BMI3D, but in this case
        we're streaming all the data so we can save it locally. So we need to tell the sink manager about
        all the channels (cortical_channels) and then hope the extractor will only use the channels that are
        actually readouts.
        '''
        super().__init__(*args, **kwargs)
        self.cortical_channels = np.arange(1,64*self.n_emg_arrays+16+8+1) # 64*n_arrays EMG + 16 AUX + 8 samples
        quattrocento.EMG.subj = self.subject_name
        quattrocento.EMG.saveid = self.saveid
        quattrocento.EMG.n_arrays = self.n_emg_arrays
        quattrocento.EMG.update_freq = self.quatt_sampling_rate

        # These get read by CorticalData when initializing the extractor
        self._neural_src_type = source.MultiChanDataSource
        self._neural_src_kwargs = dict(
            send_data_to_sink_manager=True, 
            channels=self.cortical_channels)
        self._neural_src_system_type = quattrocento.EMG

    def cleanup(self, database, saveid, **kwargs):
        '''
        Function to run at 'cleanup' time, after the FSM has finished executing. See riglib.experiment.Experiment.cleanup
        This 'cleanup' method links the file created to the database ID for the current TaskEntry
        '''
        super_result = super().cleanup(database, saveid, **kwargs)
        # Sleep time so that the plx file has time to save cleanly
        time.sleep(2)
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        filename = f'/var/tmp/tmp_{str(quattrocento.EMG)}_{self.subject_name}_{saveid}.hdf'
        print(f"Saving {filename} to database {dbname}")
        if saveid is not None:    
            if dbname == 'default':
                database.save_data(filename, "emg", saveid, True, False)
            else:
                database.save_data(filename, "emg", saveid, True, False, dbname=dbname)
        else:
            print('\n\nPlexon file not found properly! It will have to be manually linked!\n\n')

        return super_result

    @property 
    def sys_module(self):
        return quattrocento   




'''
Extensions of the render.Renderer class for stereo displays
'''

import numpy as np
from OpenGL.GL import *

from .shadow_map import ShadowMapper as Renderer
from ..utils import offaxis_frusta

class LeftRight(Renderer):
    def __init__(self, window_size, fov, near, far, focal_dist, iod, **kwargs):
        w, h = window_size
        super(LeftRight, self).__init__((w/2,h), fov, near, far, **kwargs)
        self.projections = offaxis_frusta((w/2, h), fov, near, far, focal_dist, iod)
    
    def draw(self, root, **kwargs):
        w, h = self.size
        glViewport(0, 0, w, h)
        super(LeftRight, self).draw(root, p_matrix=self.projections[0], **kwargs)
        glViewport(w, 0, w, h)
        super(LeftRight, self).draw(root, p_matrix=self.projections[1], **kwargs)

class RightLeft(LeftRight):
    def draw(self, root, **kwargs):
        w, h = self.size
        glViewport(w, 0, w, h)
        super(LeftRight, self).draw(root, p_matrix=self.projections[0], **kwargs)
        glViewport(0, 0, w, h)
        super(LeftRight, self).draw(root, p_matrix=self.projections[1], **kwargs)

class MirrorDisplay(Renderer):
    '''The mirror display requires a left-right flip, otherwise the sides are messed up'''
    def __init__(self, window_size, fov, near, far, focal_dist, iod, **kwargs):
        w, h = window_size
        flip = kwargs.pop('flip', True)
        super(MirrorDisplay, self).__init__((w/2,h), fov, near, far, **kwargs)
        self.projections = offaxis_frusta((w/2, h), fov, near, far, focal_dist, iod, flip=flip)
    
    def draw(self, root, **kwargs):
        '''
        Draw the 'root' model.

        Parameters
        ----------
        root: stereo_opengl.models.Model instance
            Root "world" model. Draws all submodels of this model.
        kwargs: optional keyword-arguments
            Optional shaders and stuff to pass to the lower-level drawing functions
        '''
        w, h = self.size
        w = int(w)
        h = int(h)

        # draw the portion of the screen with lower-left corner (0, 0), width 'w' and height 'h'
        glViewport(0, 0, w, h)
        super(MirrorDisplay, self).draw(root, p_matrix=self.projections[0], **kwargs)

        # draw the portion of the screen with lower-left corner (w, 0), width 'w' and height 'h'
        glViewport(w, 0, w, h)
        super(MirrorDisplay, self).draw(root, p_matrix=self.projections[1], **kwargs)

class DualMultisizeDisplay(Renderer):
    def __init__(self, main_window_size, mini_window_size, fov, near, far, focal_dist, iod, **kwargs):
        w, h = main_window_size
        w2, h2 = mini_window_size
        self.main_window_size = main_window_size
        self.mini_window_size = mini_window_size

        flip_main_z = kwargs.pop('flip_main_z', False)
        flip = kwargs.pop('flip', True)


        super(DualMultisizeDisplay, self).__init__((w+w2, h), fov, near, far, **kwargs)
        main_projections = offaxis_frusta(main_window_size, fov, near, far, focal_dist, iod, flip=flip, flip_z=flip_main_z)
        mini_projections = offaxis_frusta(mini_window_size, fov, near, far, focal_dist, iod, flip=flip)
        self.projections = (mini_projections[0], main_projections[0])

    def draw(self, root, **kwargs):
        '''
        Draw the 'root' model.
        
        Parameters
        ----------
        root: stereo_opengl.models.Model instance
            Root "world" model. Draws all submodels of this model.
        kwargs: optional keyword-arguments
            Optional shaders and stuff to pass to the lower-level drawing functions
        '''

        # draw the portion of the screen with lower-left corner (0, 0), width 'w' and height 'h'
        w2, h2 = self.mini_window_size
        glViewport(0, 0, w2, h2)
        super(DualMultisizeDisplay, self).draw(root, p_matrix=self.projections[0], **kwargs)

        # draw the portion of the screen with lower-left corner (w, 0), width 'w' and height 'h'
        w, h = self.main_window_size
        glViewport(w2, 0, w, h)
        super(DualMultisizeDisplay, self).draw(root, p_matrix=self.projections[1], **kwargs)

class Anaglyph(Renderer):
    '''
    From wikipedia: Anaglyph 3D is the name given to the stereoscopic 3D effect achieved by 
    means of encoding each eye's image using filters of different (usually chromatically 
    opposite) colors, typically red and cyan.
    '''
    def __init__(self, window_size, fov, near, far, focal_dist, iod, **kwargs):
        super(Anaglyph, self).__init__(window_size, fov, near, far, **kwargs)
        self.projections = offaxis_frusta(self.size, fov, near, far, focal_dist, iod)

    def draw(self, root, **kwargs):
        glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_TRUE)
        super(Anaglyph, self).draw(root, p_matrix=self.projections[0], **kwargs)
        glClear(GL_DEPTH_BUFFER_BIT)
        glColorMask(GL_FALSE, GL_TRUE, GL_TRUE, GL_TRUE)
        super(Anaglyph, self).draw(root, p_matrix=self.projections[1], **kwargs)
        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE)



# file.py
# File loading functions
import aopy
from dataclasses import dataclass
import scipy
import os
import tempfile

@dataclass
class Info():
    length: int
    units: list
    channels: list

def parse_file(filepath):
    metadata = aopy.data.load_ecube_metadata(filepath, 'Headstages')
    length = metadata['n_samples']/metadata['samplerate']
    channels = list(range(1, metadata['n_channels']+1)) # TODO incorporate the dr. map channel mapping into aopy, then use it here
    units = channels # assume multiunit exists?
    info = Info(length, units, channels)
    return info

def load_lfp(filepath, samplerate=1000):
    h5file = tempfile.NamedTemporaryFile(suffix=".hdf", delete=False)
    h5file.flush()
    h5file.close()
    tmp_filepath = h5file.name
    # tmp_filepath = '/home/pagaiisland/leo/temp_file.hdf'
    files = {
        'ecube': filepath,
    }
    aopy.preproc.proc_lfp('', files, '', tmp_filepath, overwrite=True)
    import h5py
    file = h5py.File(tmp_filepath, 'r')
    lfp_data = file['lfp_data']
    # broadband_data = aopy.data.load_ecube_data(filepath, 'Headstages')
    
    # TODO write downsampling code in aopy!!!! this is untested!!!!
    # metadata = aopy.data.load_ecube_metadata(filepath, 'Headstages')
    # bb_samples = metadata['n_samples']
    # lfp_samples = int(bb_samples/(metadata['samplerate']/samplerate))
    # lfp_data = scipy.signal.resample(broadband_data, lfp_samples, axis=0)
    return lfp_data

def load_bmi3d_cycle_times(files):
    data, metadata = aopy.preproc.parse_bmi3d("", files)
    times = data['clock']['timestamp_sync']
    return times



"""
Quaternion class.

@author: Peter Corke
@copyright: Peter Corke 1999-2008
"""

from numpy import *
from .utility import *
from . import transform as T
import copy

# Copyright (C) 1999-2002, by Peter I. Corke

class Quaternion:
    """A quaternion is a compact method of representing a 3D rotation that has
    computational advantages including speed and numerical robustness.

    A quaternion has 2 parts, a scalar s, and a vector v and is typically written::

        q = s <vx vy vz>

    A unit quaternion is one for which M{s^2+vx^2+vy^2+vz^2 = 1}.

    A quaternion can be considered as a rotation about a vector in space where
    q = cos (theta/2) sin(theta/2) <vx vy vz>
    where <vx vy vz> is a unit vector.

    Various functions such as INV, NORM, UNIT and PLOT are overloaded for
    quaternion objects.

    Arithmetic operators are also overloaded to allow quaternion multiplication,
    division, exponentiaton, and quaternion-vector multiplication (rotation).
    """

    def __init__(self, *args):
        '''
Constructor for quaternion objects:

    - q = Quaternion(v, theta)    from vector plus angle
    - q = Quaternion(R)       from a 3x3 or 4x4 matrix
    - q = Quaternion(q)       from another quaternion
    - q = Quaternion(s)       from a scalar
    - q = Quaternion(v)       from a matrix/array/list v = [s v1 v2 v3]
    - q = Quaternion(s, v1, v2, v3)    from 4 elements
    - q = Quaternion(s, v)    from 4 elements
'''

        self.vec = [];
        
        if len(args) == 0:
                # default is a null rotation
                self.s = 1.0
                self.v = matrix([0.0, 0.0, 0.0])

        elif len(args) == 1:
            arg = args[0]
            
            if isinstance(arg, Quaternion):
            # Q = QUATERNION(q) from another quaternion
                self.s = arg.s
                self.v = arg.v
                return
                
            if type(arg) is matrix:
                # Q = QUATERNION(R) from a 3x3
                if (arg.shape == (3,3)):
                    Q = Quaternion.tr2q(arg);
                    self.s = Q.s
                    self.v = Q.v
                    return 

                # Q = QUATERNION(R) from a 4x4
                if (arg.shape == (4,4)):
                    Q = Quaternion.tr2q(arg[0:3,0:3]);
                    self.s = Q.s
                    self.v = Q.v
                    
                    return 
            # some kind of list, vector, scalar...
            
            v = arg2array(arg);
            
            if len(v) == 4:
                # Q = QUATERNION([s v1 v2 v3]) from 4 elements
                self.s = v[0];
                self.v = mat(v[1:4]);
                
            elif len(v) == 3:
                self.s = 0
                self.v = mat(v[0:3]);

            elif len(v) == 1:
                # Q = QUATERNION(s) from a scalar
                #Q = QUATERNION(s)               from a scalar
                self.s = v[0];
                self.v = mat([0, 0, 0])

        elif len(args) == 2:
            # Q = QUATERNION(v, theta) from vector plus angle
            # Q = Quaternion(s, v);

            a1 = arg2array(args[0]);
            a2 = arg2array(args[1]);
            
            if len(a1) == 1 and len(a2) == 3:
                # s, v
                self.s = a1[0];
                self.v = mat(a2);
            elif len(a1) == 3 and len(a2) == 1:
                # v, theta
                self.s = a2[0];
                self.v = mat(a1);

        elif len(args) == 4:
            self.s = args[0];
            self.v = mat(args[1:4])

        else:
                print("error")
                return None

    def __repr__(self):
            return "%f <%f, %f, %f>" % (self.s, self.v[0,0], self.v[0,1], self.v[0,2])


    @staticmethod
    def tr2q(t):
        #TR2Q   Convert homogeneous transform to a unit-quaternion
        #
        #   Q = tr2q(T)
        #
        #   Return a unit quaternion corresponding to the rotational part of the
        #   homogeneous transform T.
        Q = Quaternion()
        qs = sqrt(trace(t)+1)/2.0
        kx = t[2,1] - t[1,2]    # Oz - Ay
        ky = t[0,2] - t[2,0]    # Ax - Nz
        kz = t[1,0] - t[0,1]    # Ny - Ox

        if (t[0,0] >= t[1,1]) and (t[0,0] >= t[2,2]):
                kx1 = t[0,0] - t[1,1] - t[2,2] + 1      # Nx - Oy - Az + 1
                ky1 = t[1,0] + t[0,1]           # Ny + Ox
                kz1 = t[2,0] + t[0,2]           # Nz + Ax
                add = (kx >= 0)
        elif (t[1,1] >= t[2,2]):
                kx1 = t[1,0] + t[0,1]           # Ny + Ox
                ky1 = t[1,1] - t[0,0] - t[2,2] + 1  # Oy - Nx - Az + 1
                kz1 = t[2,1] + t[1,2]           # Oz + Ay
                add = (ky >= 0)
        else:
                kx1 = t[2,0] + t[0,2]           # Nz + Ax
                ky1 = t[2,1] + t[1,2]           # Oz + Ay
                kz1 = t[2,2] - t[0,0] - t[1,1] + 1  # Az - Nx - Oy + 1
                add = (kz >= 0)

        if add:
                kx = kx + kx1
                ky = ky + ky1
                kz = kz + kz1
        else:
                kx = kx - kx1
                ky = ky - ky1
                kz = kz - kz1

        kv = matrix([kx, ky, kz])
        nm = linalg.norm( kv )
        if nm == 0:
                Q.s = 1.0
                Q.v = matrix([0.0, 0.0, 0.0])

        else:
                Q.s = qs
                Q.v = (sqrt(1 - qs**2) / nm) * kv
        
        return Q

    ############### OPERATORS #########################################
    #PLUS Add two quaternion objects
    #
    # Invoked by the + operator
    #
    # q1+q2 standard quaternion addition
    def __add__(self, q):
        '''
        Return a new quaternion that is the element-wise sum of the operands.
        '''
        if isinstance(q, Quaternion):
            qr = Quaternion()
            qr.s = 0

            qr.s = self.s + q.s
            qr.v = self.v + q.v

            return qr
        else:
            raise ValueError

    #MINUS Subtract two quaternion objects
    #
    # Invoked by the - operator
    #
    # q1-q2 standard quaternion subtraction

    def __sub__(self, q):
        '''
        Return a new quaternion that is the element-wise difference of the operands.
        '''
        if isinstance(q, Quaternion):
            qr = Quaternion()
            qr.s = 0

            qr.s = self.s - q.s
            qr.v = self.v - q.v

            return qr
        else:
            raise ValueError

    # q * q  or q * const
    def __mul__(self, q2):
        '''
        Quaternion product. Several cases are handled
        
            - q * q   quaternion multiplication
            - q * c   element-wise multiplication by constant
            - q * v   quaternion-vector multiplication q * v * q.inv();
        '''
        qr = Quaternion();
        
        if isinstance(q2, Quaternion):
                
            #Multiply unit-quaternion by unit-quaternion
            #
            #   QQ = qqmul(Q1, Q2)

            # decompose into scalar and vector components
            s1 = self.s;    v1 = self.v
            s2 = q2.s;  v2 = q2.v

            # form the product
            qr.s = s1*s2 - v1*v2.T
            qr.v = s1*v2 + s2*v1 + cross(v1,v2)

        elif type(q2) is matrix:
                
            # Multiply vector by unit-quaternion
            #
            #   Rotate the vector V by the unit-quaternion Q.

            if q2.shape == (1,3) or q2.shape == (3,1):
                    qr = self * Quaternion(q2) * self.inv()
                    return qr.v;
            else:
                    raise ValueError;

        else:
            qr.s = self.s * q2
            qr.v = self.v * q2

        return qr

    def __rmul__(self, c):
        '''
        Quaternion product. Several cases are handled
 
            - c * q   element-wise multiplication by constant
        '''
        qr = Quaternion()
        qr.s = self.s * c
        qr.v = self.v * c

        return qr
        
    def __imul__(self, x):
        '''
        Quaternion in-place multiplication
        
            - q *= q2
            
        '''
        
        if isinstance(x, Quaternion):
            s1 = self.s;   
            v1 = self.v
            s2 = x.s
            v2 = x.v

            # form the product
            self.s = s1*s2 - v1*v2.T
            self.v = s1*v2 + s2*v1 + cross(v1,v2)

        elif isscalar(x):
            self.s *= x;
            self.v *= x;

        return self;


    def __div__(self, q):
        '''Return quaternion quotient.  Several cases handled:
            - q1 / q2      quaternion division implemented as q1 * q2.inv()
            - q1 / c       element-wise division
        '''
        if isinstance(q, Quaternion):
            qr = Quaternion()
            qr = self * q.inv()
        elif isscalar(q):
            qr.s = self.s / q
            qr.v = self.v / q

        return qr


    def __pow__(self, p):
        '''
        Quaternion exponentiation.  Only integer exponents are handled.  Negative
        integer exponents are supported.
        '''
        
        # check that exponent is an integer
        if not isinstance(p, int):
            raise ValueError
        
        qr = Quaternion()
        q = Quaternion(self);
        
        # multiply by itself so many times
        for i in range(0, abs(p)):
            qr *= q

        # if exponent was negative, invert it
        if p < 0:
            qr = qr.inv()

        return qr

    def copy(self):
        """
        Return a copy of the quaternion.
        """
        return copy.copy(self);
                
    def inv(self):
        """Return the inverse.
        
        @rtype: quaternion
        @return: the inverse
        """
        
        qi = Quaternion(self);
        qi.v = -qi.v;
        
        return qi;



    def norm(self):
        """Return the norm of this quaternion.
        
        @rtype: number
        @return: the norm
        """
        
        return linalg.norm(self.double())

    def double(self):
        """Return the quaternion as 4-element vector.
        
        @rtype: 4-vector
        @return: the quaternion elements
        """
        return concatenate( (mat(self.s), self.v), 1 )


    def unit(self):
        """Return an equivalent unit quaternion
        
        @rtype: quaternion
        @return: equivalent unit quaternion
        """
        
        qr = Quaternion()
        nm = self.norm()

        qr.s = self.s / nm
        qr.v = self.v / nm

        return qr


    def T(self):
        """Return an equivalent rotation matrix.
        
        @rtype: 4x4 homogeneous transform
        @return: equivalent rotation matrix
        """

        return T.r2t( self.R() )

    def R(self):
        """Return an equivalent rotation matrix.
        
        @rtype: 3x3 orthonormal rotation matrix
        @return: equivalent rotation matrix
        """

        s = self.s;
        x = self.v[0,0]
        y = self.v[0,1]
        z = self.v[0,2]

        return matrix([[ 1-2*(y**2+z**2),   2*(x*y-s*z),    2*(x*z+s*y)],
                       [2*(x*y+s*z),    1-2*(x**2+z**2),    2*(y*z-s*x)],
                       [2*(x*z-s*y),    2*(y*z+s*x),    1-2*(x**2+y**2)]])

    

#QINTERP Interpolate rotations expressed by quaternion objects
#
#   QI = qinterp(Q1, Q2, R)
#
# Return a unit-quaternion that interpolates between Q1 and Q2 as R moves
# from 0 to 1.  This is a spherical linear interpolation (slerp) that can
# be interpretted as interpolation along a great circle arc on a sphere.
#
# If r is a vector, QI, is a cell array of quaternions, each element
# corresponding to sequential elements of R.
#
# See also: CTRAJ, QUATERNION.

# MOD HISTORY
# 2/99 convert to use of objects
# $Log: qinterp.m,v $
# Revision 1.3  2002/04/14 11:02:54  pic
# Changed see also line.
#
# Revision 1.2  2002/04/01 12:06:48  pic
# General tidyup, help comments, copyright, see also, RCS keys.
#
# $Revision: 1.3 $
#
# Copyright (C) 1999-2002, by Peter I. Corke

    def interp(Q1, Q2, r):
        q1 = Q1.double()
        q2 = Q2.double()

        theta = arccos(q1*q2.T)
        q = []
        count = 0

        if isscalar(r):
            if r<0 or r>1:
                raise Exception('R out of range')
            if theta == 0:
                q = Quaternion(Q1)
            else:
                q = Quaternion( (sin((1-r)*theta) * q1 + sin(r*theta) * q2) / sin(theta) )
        else:
            for R in r:
                if theta == 0:
                    qq = Q1
                else:
                    qq = Quaternion( (sin((1-R)*theta) * q1 + sin(R*theta) * q2) / sin(theta) )
                q.append(qq)
        return q




'''Needs docs'''


import os
import operator
import numpy as np
from OpenGL.GL import *

from ..utils import perspective, orthographic
from .shader import ShaderProgram

cwd = os.path.join(os.path.abspath(os.path.split(__file__)[0]), "..")

class _textrack(object):
    pass

class Renderer(object):
    def __init__(self, window_size, fov, near, far, modelview=None, shaders=None, programs=None, light_direction=None):
        self.render_queue = None
        self.size = window_size
        self.drawpos = 0,0
        w, h = window_size
        self.projection = perspective(fov, w / h, near, far)
        self.light_direction = light_direction or (-1., -2., -2., 0.)
        self.modelview = np.eye(4) if modelview is None else modelview

        #Add the default shaders
        if shaders is None:
            shaders = dict()
            shaders['passthru'] = GL_VERTEX_SHADER, "passthrough.v.glsl"
            shaders['default'] = GL_FRAGMENT_SHADER, "default.f.glsl", "phong.f.glsl"
            shaders['ui_passthru'] = GL_VERTEX_SHADER, "passthrough2d.v.glsl"
            shaders['ui_default'] = GL_FRAGMENT_SHADER, "none.f.glsl"
        if programs is None:
            programs = dict()
            programs['default'] = "passthru", "default"
            programs['ui'] = "ui_passthru", "ui_default"

        #compile the given shaders and the programs
        self.shaders = dict()
        for k, v in list(shaders.items()):
            # print("Compiling shader %s..."%k)
            self.add_shader(k, *v)
        
        self.programs = dict()
        for name, shaders in list(programs.items()):
            self.add_program(name, shaders)
        
        #Set up the texture units
        self.reset_texunits()

        #Generate the default fullscreen quad
        verts = np.array([(-1,-1,0,1), (1,-1,0,1), (1,1,0,1), (-1,1,0,1)]).astype(np.float32)
        polys = np.array([(0,1,2),(0,2,3)]).astype(np.uint16)
        vbuf = glGenBuffers(1)
        ebuf = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, vbuf)
        glBufferData(GL_ARRAY_BUFFER, verts, GL_STATIC_DRAW)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, polys, GL_STATIC_DRAW)
        self.fsquad_buf = vbuf, ebuf
    
    def _queue_render(self, root, shader=None):
        '''

        Parameters
        ----------
        root: stereo_opengl.models.Model instance
            The root model from which to start drawing
        shader: ???????, default=None
            ????????
        '''
        queue = dict((k, dict()) for k in list(self.programs.keys()))

        for pname, drawfunc, tex in root.render_queue(shader=shader):
            if tex not in queue[pname]:
                queue[pname][tex] = []
            queue[pname][tex].append(drawfunc)
        
        for pname in list(self.programs.keys()):
            #assert len(self.texavail) > len(queue[pname])
            for tex in list(queue[pname].keys()):
                if tex is not None:
                    self.get_texunit(tex)
        
        self.render_queue = queue
    
    def get_texunit(self, tex):
        '''Input a Texture object, output a tuple (index, TexUnit)'''
        if tex not in self.texunits:
            unit = self.texavail.pop()
            glActiveTexture(unit[1])
            if tex == "None":
                glBindTexture(GL_TEXTURE_2D, 0)
            else:
                glBindTexture(GL_TEXTURE_2D, tex.tex)
            #print "Binding %r to %d"%(tex, unit[0])
            self.texunits[tex] = unit[0]
        
        return self.texunits[tex]
    
    def reset_texunits(self):
        maxtex = glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS)
        # print(f"Max texture units: {maxtex}")
        # glActiveTexture(GL_TEXTURE0)  # Reset to default texture unit
        self.texavail = set((i, globals()['GL_TEXTURE%d'%i]) for i in range(1, maxtex))
        self.texunits = dict() 

    def add_shader(self, name, stype, filename, *includes):
        src = []
        main_path = os.path.join(cwd, "shaders", filename)
        with open(main_path, 'r') as main:
            version = main.readline().strip()
            main_content = main.read()
        
        for inc in includes:
            inc_path = os.path.join(cwd, "shaders", inc)
            with open(inc_path, 'r') as incfile:
                ver = incfile.readline().strip()
                assert ver == version, f"Version mismatch: {ver} != {version}"
                src.append(incfile.read())
        
        src.append(main_content)
        
        full_src = "\n".join([version] + src)
        shader = glCreateShader(stype)
        glShaderSource(shader, full_src)
        glCompileShader(shader)

        if not glGetShaderiv(shader, GL_COMPILE_STATUS):
            err = glGetShaderInfoLog(shader)
            glDeleteShader(shader)
            print(f"Shader compilation error for {filename}:")  # Debug print
            print(err.decode('utf-8'))  # Debug print
            raise Exception(err)
        
        self.shaders[name] = shader
    
    def add_program(self, name, shaders):
        shaders = [self.shaders[i] for i in shaders]
        sp = ShaderProgram(shaders)
        self.programs[name] = sp
    
    def draw(self, root, shader=None, apply_default=False, requeue=False, **kwargs):
        if self.render_queue is None or requeue:
            self._queue_render(root)
        
        if "p_matrix" not in kwargs:
            kwargs['p_matrix'] = self.projection
        if "modelview" not in kwargs:
            kwargs['modelview'] = self.modelview
        if "light_direction" not in kwargs:
            kwargs['light_direction'] = self.light_direction
        
        if shader is not None:
            if apply_default: # apply this shader to all models that don't have a program specified
                self.programs[shader].draw(self, self.render_queue['default'], **kwargs)
            else: # apply this shader to only the models that have this shader specified
                self.programs[shader].draw(self, self.render_queue[shader], **kwargs)
        else:
            for name, program in list(self.programs.items()):
                program.draw(self, self.render_queue[name], **kwargs)

        error = glGetError()
        if error != GL_NO_ERROR:
            print(f"OpenGL Error: {error}")
    
    def draw_done(self):
        self.reset_texunits()

class Renderer2D(Renderer):

    def __init__(self, screen_cm):
        shaders = dict()
        shaders['passthru'] = GL_VERTEX_SHADER, "passthrough2d.v.glsl"
        shaders['default'] = GL_FRAGMENT_SHADER, "none.f.glsl"
        programs = dict()
        programs['default'] = "passthru", "default"
        programs['ui'] = "passthru", "default"
        super().__init__(screen_cm, np.nan, 1, 1024, shaders=shaders, programs=programs)
        w, h = screen_cm
        self.projection = orthographic(w, h, 1, 1024)

def test():
    import pygame
    pygame.init()
    pygame.display.set_mode((100,100), pygame.OPENGL | pygame.DOUBLEBUF)

    return Renderer(
                shaders=dict(
                    passthru=(GL_VERTEX_SHADER, "passthrough.v.glsl"),
                    phong=(GL_FRAGMENT_SHADER, "default.f.glsl", "phong.f.glsl")), 
                programs=dict(
                    default=("passthru", "phong"),
                )
            )


import socket
import time
import numpy as np
import pandas as pd

class IncompleteSampleError(Exception):
    def __init__(self, actual_sample, expected_sample):
        super().__init__()
        self.error_message = \
            'Incomplete sample. Received: {}. Expected: {}.'.format(
                actual_sample, expected_sample)


class Quattrocento():
    """

    Reference EMG device for high-throughput recordings.
    The user manual for the device can be found [here](
    https://www.otbioelettronica.it/attachments/article/70/
    Quattrocento%20User%20Manual%20v1.4.pdf)
    The user manual for the OTBioLab software can be found [here](https://
    www.otbiolab.it/UserManuals/User%20Manual%20OT%20BioLab+%20v1.0.pdf)
    This class will connect to the Quattrocento using a raw TCP network socket
    (see host and port defaults below), then send a configuration buffer using
    the protocol described in https://www.otbioelettronica.it/en/?preview=1&op
    tion=com_dropfiles&format&task=frontfile.download&catid=41&id=70&Itemid=10
    00000000000.
    Among other things, configuration includes the sampling rate and the number
    of electrode arrays (at 96 channels per array) to use.
    When streaming is activated, the Quattrocento is expected to send a
    continuous series of data buffers, each representing a single multi-
    channel sample and structured as contiguous 16-bit little-endian integers.
        EMG: main EMG array channels as contiguous 16-bit little-endian ints
        Aux: 16 auxiliary channels as 16-bit little-endian ints
        Acc: 8 accessory channels as 16-bit
    For example, if this reader is configured with num_arrays=2 (acquisition
    with two out of the possible four electrode arrays active) it will expect
    to receive data buffers exactly 432 bytes in length:
        EMG: 2 arrays * 96 channels per array * 2B per channel = 384 bytes
        Aux: 16 auxiliary channels * 2B per channel = 32 bytes
        Acc: 8 accessory channels * 2B per channel = 16 bytes
    With all four arrays active, each sample buffer will be 816 bytes.
    At present, the auxiliary channels are unused, and only the first
    accessory channel (sample number) is used.
    Other than the expectation that sample_num is monotonic, there are no
    documented framing or checksum structures to assist with finding buffer
    boundaries in the TCP stream. We must know the message size exactly split
    the stream at the appropriate points.
    Note that electrode numbering within the EMG segment is determined by the
    geometry of the surface electrodes used for the recording.
    """

    TIMEOUT = 5.0  # warn if no data received after this many seconds

    RECEIVE_LOOP_PERIOD = 0.05# 100 ms of data

    # communication with Quattrocento device directly
    DIRECT_PORT = 23456
    DIRECT_HOST = '169.254.1.10'

    CHANNELS_PER_ARRAY = 64
    #CHANNELS_PER_ARRAY = (384 + 16 + 8)
    ACCESSORY_CHANNELS = 16
    EXTENDED_CHANNELS = 8

    valid_stream_names = ["quattrocento", "main"]

    df_data_output = pd.DataFrame()

    def __init__(self,
                 f_samp: int = 2048,
                 num_arrays: int = 1,
                 name: str = 'Quattrocento',
                 host: str = DIRECT_HOST,
                 port: int = DIRECT_PORT,
                 emg_output: str = 'temp.npy',
                 data_output: str = 'tmp.bufer'
                 ) -> None:
        self.port = port
        self.host = host
        self.timer = name
        self._last_data_timestamp = time.time()
        self.f_samp = f_samp
        self.name = name
        self.num_arrays = num_arrays
        self.num_emg_channels = self.num_arrays * self.CHANNELS_PER_ARRAY
        self.num_sample_fields = self.num_emg_channels + \
            self.ACCESSORY_CHANNELS + self.EXTENDED_CHANNELS
        self.sample_message_size = self.num_sample_fields * 2  # 2B per field

        self.socket: socket.socket
        self.ready = False
        self.receiving = False
        self.quatt_driver = QuattrocentoDriver()

        self.emg_output = emg_output
        self.data_output = data_output

    def teardown(self):
        stop_acq_str = self.quatt_driver.create_config(
            False, False, self.f_samp, self.num_arrays)
        self.socket.send(stop_acq_str)

        self.socket.close()

    def setup(self):
        server_addr = (self.host, self.port)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.settimeout(None)  # setting this to None is blocking
            self.socket.connect(server_addr)
        except Exception as e:
            print("Quattrocento: error: {}".format(e))
            raise
        print(
            "Quattrocento: {} receiver subscribing to socket {}...".format(
                self.name, server_addr))

        self.ready = True

    def _receive_data(self):
        # print("inside _receive_data..")
        # for i in range(2):

        sample_nums = []
        for i in range(int(self.f_samp * self.RECEIVE_LOOP_PERIOD)):
            try:
                # recv waits until buffer of size self.num_channels*2 is full
                data = self.socket.recv(self.sample_message_size,
                                        socket.MSG_WAITALL)  # bytes to read
                # data = self.socket.recv(self.sample_message_size)  # bytes to read

                if len(data) < (self.sample_message_size):
                    raise IncompleteSampleError(
                        len(data), self.num_sample_fields * 2)

            except IncompleteSampleError as e:
                print(f'Quattrocento: {e.error_message}')
                return

            except OSError as e:
                print("Quattrocento: error: {}".format(e))
                data_lag = time.time() - self._last_data_timestamp
                if data_lag >= self.TIMEOUT:
                    print(
                        f"Quattrocento: No data received from {self.name} in"
                        " the last {self.TIMEOUT} seconds.")
                    self._last_data_timestamp = time.time()
                return

            self._last_data_timestamp = time.time()
            emg, sample_num = self._decode_message(data)
            # print("emg: ", type(emg))

        # ### SAVING FUNCTIONS ####
        # with open(self.emg_output, "ab") as f:
        #     np.savetxt(f, [emg], delimiter=',')

        # # https://stackoverflow.com/questions/30376581/save-numpy-array-in-append-mode
        # print(f'Saving to {self.data_output}')
        # f = open(self.data_output, 'ab') # 'tmp.bufer'
        # f.write(data)
        # f.close()

        return emg, sample_nums


    # written by Gus for labgraph
    def get_state(self):
        print("inside get_state..")
        # for i in range(2):
        for i in range(int(self.f_samp * self.RECEIVE_LOOP_PERIOD)):
            print("int i: ", i)
            try:
                # recv waits until buffer of size self.num_channels*2 is full
                data = self.socket.recv(self.sample_message_size,
                                        socket.MSG_WAITALL)  # bytes to read
                # data = self.socket.recv(self.sample_message_size)  # bytes to read

                if len(data) < (self.sample_message_size):
                    raise IncompleteSampleError(
                        len(data), self.num_channels * 2)

            except IncompleteSampleError as e:
                print(f'Quattrocento: {e.error_message}')
                return

            except OSError as e:
                print("Quattrocento: error: {}".format(e))
                data_lag = time.time() - self._last_data_timestamp
                if data_lag >= self.TIMEOUT:
                    print(
                        f"Quattrocento: No data received from {self.name} in"
                        " the last {self.TIMEOUT} seconds.")
                    self._last_data_timestamp = time.time()
                return

            self._last_data_timestamp = time.time()
            emg, sample_num = self._decode_message(data)
            print("returning emg")
            return emg, sample_num


    def start_acq(self):
        start_acq_str = self.quatt_driver.create_config(
            True, False, self.f_samp, self.num_arrays)
        self.socket.send(start_acq_str)

    def stream_transform(self):
        if self.ready and not self.receiving:
            self.start_acq()
            self.receiving = True
        # print("starting receive data")
        emg, sample_num = self._receive_data()
        return emg,sample_num

    # for use in calling from labgraph
    def start_stream(self):
        if self.ready and not self.receiving:
            self.start_acq()
            self.receiving = True
       

    def _decode_message(self, message):
        # Convert the EMG section directly into a Numpy array using frombuffer
        # with int16, little endian arrangement
        dt = np.dtype('int16')
        dt = dt.newbyteorder('<')
        emg = np.frombuffer(message, dtype=dt)
        # Pick out the sample number, which is the first item in the extended
        # channel section, and must be interpreted as an unsigned short.
        #sample_num = struct.unpack('<H', message[-16:-14])[0]
        print(emg.shape)
        sample_num = emg[-self.EXTENDED_CHANNELS]

        return emg[:self.num_emg_channels], sample_num

    def _output(self, emg, sample_num, timestamp):
        sample = {
            "TIME": timestamp,
            "MACHINE_TIME": time.time(),
            "EMG": emg,
            "SAMPLE": sample_num
        }


class QuattrocentoDriver():
    """Driver for interfacing with Quattrocento device. The driver was
    designed based on the configuration protocol from the manufacturer,
    which can be found [here](https://www.otbioelettronica.it/en/
    ?preview=1&option=com_dropfiles&format=&task=frontfile.download
    &catid=41&id=70&Itemid=1000000000000)
    """

    # BYTE 1: ACQ_SETT
    ACQ_OFF = np.uint8(0b00000000)
    ACQ_ON = np.uint8(0b00000001)

    ALL_INPUTS_ACTIVE = np.uint8(0b00000110)
    THREE_FOURTHS_INPUTS_ACTIVE = np.uint8(0b00000100)
    HALF_INPUTS_ACTIVE = np.uint8(0b00000010)
    ONE_FOURTH_INPUTS_ACTIVE = np.uint8(0b00000000)

    SAMPLING_FREQ_512 = np.uint8(0b00000000)
    SAMPLING_FREQ_2048 = np.uint8(0b00001000)
    SAMPLING_FREQ_5120 = np.uint8(0b00010000)
    SAMPLING_FREQ_10240 = np.uint8(0b00011000)

    RECORDING_ON = np.uint8(0b00100000)
    RECORDING_OFF = np.uint8(0b00000000)

    DECIMATOR_ON = np.uint8(0b01000000)
    DECIMATOR_OFF = np.uint8(0b00000000)

    # BYTES 2 and 3: analog outputs
    AN_OUT_IN_SEL = np.uint8(0b00000000)
    AN_OUT_CH_SEL = np.uint8(0b00000000)

    # BYTES 4 - 39: configure 8, 16-chnl IN & 4, 64-chnl MULTIPLE_IN ports
    IN_CONFIG_0 = np.uint8(0b00000000)

    CONFIG_1_ADAPTER_64_CHANNEL = np.uint8(0b00000100)

    CONFIG_2_SIDE_NOT_DEF = np.uint8(0b00000000)
    CONFIG_2_HPF_03HZ = np.uint8(0b00000000)
    CONFIG_2_HPF_10HZ = np.uint8(0b00010000)
    CONFIG_2_LPF_4400HZ = np.uint8(0b00001100)
    CONFIG_2_LPF_900HZ = np.uint8(0b00001000)
    CONFIG_2_MODE_MONO = np.uint8(0b00000000)
    CONFIG_2_MODE_BI = np.uint8(0b00000010)
    CONFIG_2_MODE_DIFF = np.uint8(0b00000001)

    # CRC CHECKSUM
    def crc8(self, vals):
        crc = np.uint8(0)
        for v in vals:
            for j in range(8):
                s = np.not_equal(np.mod(crc, 2), np.mod(v, 2))
                crc = np.uint8(np.floor_divide(crc, 2))
                if s:
                    crc = np.uint8(np.bitwise_xor(crc, 140))
                v = np.uint8(np.floor_divide(v, 2))
        return crc

    def create_config(self, start_acq, start_rec, f_samp, num_arrays):
        if start_acq:
            acq = self.ACQ_ON
        else:
            acq = self.ACQ_OFF

        if start_rec:
            rec = self.RECORDING_ON
        else:
            rec = self.RECORDING_OFF

        num_arrays_to_inputs = {
            1: self.ONE_FOURTH_INPUTS_ACTIVE,
            2: self.HALF_INPUTS_ACTIVE,
            3: self.THREE_FOURTHS_INPUTS_ACTIVE,
            4: self.ALL_INPUTS_ACTIVE
        }

        inputs = num_arrays_to_inputs[num_arrays]

        sampling_rate_selector = {
            512: self.SAMPLING_FREQ_512,
            2048: self.SAMPLING_FREQ_2048,
            5120: self.SAMPLING_FREQ_5120,
            10240: self.SAMPLING_FREQ_10240
        }

        sampling_rate = sampling_rate_selector[f_samp]

        acq_sett_list = [
            np.uint8(0b10000000), acq, inputs, sampling_rate, rec,
            self.DECIMATOR_OFF
        ]
        ACQ_SETT = np.bitwise_or.reduce(acq_sett_list)
        config_list = np.array(ACQ_SETT, dtype=np.uint8)

        config_list = np.append(config_list, self.AN_OUT_IN_SEL)
        config_list = np.append(config_list, self.AN_OUT_CH_SEL)

        # Set up input ports
        IN_CONFIG_1_ON = self.CONFIG_1_ADAPTER_64_CHANNEL
        IN_CONFIG_1_OFF = np.uint8(0b00000000)

        config_2_list = [
            self.CONFIG_2_SIDE_NOT_DEF, self.CONFIG_2_HPF_10HZ,
            self.CONFIG_2_LPF_900HZ, self.CONFIG_2_MODE_MONO
        ]

        IN_CONFIG_2_ON = np.bitwise_or.reduce(config_2_list)
        IN_CONFIG_2_OFF = np.uint8(0b00000000)

        # Set up 8 IN ports
        for i in range(num_arrays * 2):
            config_list = np.append(config_list, self.IN_CONFIG_0)
            config_list = np.append(config_list, IN_CONFIG_1_ON)
            config_list = np.append(config_list, IN_CONFIG_2_ON)

        for i in range(8 - (num_arrays * 2)):
            config_list = np.append(config_list, self.IN_CONFIG_0)
            config_list = np.append(config_list, IN_CONFIG_1_OFF)
            config_list = np.append(config_list, IN_CONFIG_2_OFF)

        # Set up 4 MULTIPLE IN ports
        for i in range(num_arrays):
            config_list = np.append(config_list, self.IN_CONFIG_0)
            config_list = np.append(config_list, IN_CONFIG_1_ON)
            config_list = np.append(config_list, IN_CONFIG_2_ON)

        for i in range(4 - num_arrays):
            config_list = np.append(config_list, self.IN_CONFIG_0)
            config_list = np.append(config_list, IN_CONFIG_1_OFF)
            config_list = np.append(config_list, IN_CONFIG_2_OFF)

        crc = self.crc8(config_list)
        config_list = np.append(config_list, crc)
        config_list = config_list.tobytes()

        return config_list


class QuattroOtlight:    

    emg_channels = 384
    aux_channels = 16
    accessory_channels = 8

    def __init__(self, host='127.0.0.1', port=31000, emg_ch_range=[16*8,16*8+64], refresh_freq=1, sample_freq=2048):
        '''
        Choose the same refresh frequency and sampling frequency as the settings in OTBioLabLight. The EMG channel range
        describes the location of the electrodes in use. Default is MULTI IN 1 channels.

        Parameters
        ----------
        Host should be set to the computer running the OTBioLab Light
        Refresh rate should be set as high as possible (32) and match what is set in OTBioLab
        Sample rate should be set to match what is set in OTBioLab
        '''
        self.host = host
        self.port = port
        self.emg_ch_range = emg_ch_range
        self.refresh_freq = refresh_freq
        self.sample_freq = sample_freq

        self.sample_message_size = self.emg_channels + self.aux_channels + self.accessory_channels
        self.sample_message_size_in_bytes = int(self.sample_message_size * 2) # two bytes per number
        self.batch_size = int(self.sample_freq // self.refresh_freq)

    def setup(self):
        server_addr = (self.host, self.port)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.settimeout(None)  # setting this to None is blocking
            self.socket.connect(server_addr)
        except Exception as e:
            print("Quattrocento: error: {}".format(e))
            raise
        print(
            "Quattrocento: {} receiver subscribing to socket {}...".format(
                'quattro', server_addr))

        self.socket.send('startTX'.encode())
        data = self.socket.recv(8).decode() # read in 8 bytes to confirm connection
        if data != 'OTBioLab':
            raise Exception(f'cannot connect to the device at {server_addr}')
        
        print(f'connected to the device ')
        time.sleep(0.1)

    def read_emg(self):
        """
        read in a batch of emg  and decode into a tuple of 
        (emg_channels, aux_channels, sample_counter)
        """
        data = self.socket.recv(self.sample_message_size_in_bytes *self.batch_size,
                                socket.MSG_WAITALL )  # bytes to read
        dt = np.dtype('int16')
        dt = dt.newbyteorder('<')
        # data.shape here is (64, 408) - 408 = 384 (emg channels) + 16 (aux) + 8 (accessory)
        data = np.frombuffer(data, dtype=dt).reshape((-1, self.sample_message_size,))
        # data is (64, 408) - so datapoints x num channels
        
        return self._decode_emg(data)


    def _decode_emg(self, data:np.ndarray):
        """
        decompose data into actual emg, aux and accessory channels
        data is num_data_points by num_channels
        """
        emg_signals  = data[:, :self.emg_channels]
        aux_signals = data[:, self.emg_channels:self.emg_channels+self.aux_channels]
        sample_counter = data[:, -self.accessory_channels:]

        # select only the output channels
        emg_signals = emg_signals[:, self.emg_ch_range[0]:self.emg_ch_range[1]]

        return (emg_signals, aux_signals, sample_counter)

    def tear_down(self):
        self.socket.send('stopTX'.encode())

        self.socket.close()

class QuattroDummy(QuattroOtlight):
    
    supported_dummy_data_types = ['zeros','random']
    
    def __init__(self, chan_range = None, dummy_data_type = 'zeros') -> None:
        
        super().__init__()
        
        if dummy_data_type in self.supported_dummy_data_types: 
            self.dumm_data_type = dummy_data_type
        else:
            raise Exception(f'{dummy_data_type} is not in supported dummy data types {self.supported_dummy_data_types}')
        

    def setup(self):
        pass

    def read_emg(self):
        """
        read in a batch of emg  and decode into a tuple of 
        (emg_channels, aux_channels, sample_counter)
        """
        start_time = time.time()
        
        if self.dumm_data_type == "zeros":
            data = np.zeros((self.batch_size, self.sample_message_size))
        elif self.dumm_data_type == "random":
            data = np.random.rand(self.batch_size, self.sample_message_size)
        else:
            raise NotImplementedError(f'has not implemented for dummy_data_type {self.dumm_data_type}')
        

        decoded_data = self._decode_emg(data)
        end_time = time.time()
        loop_time = end_time - start_time
        
        # empirically subtract 0.01 to ensure the actual refresh_freq = 32 Hz
        time.sleep(1/self.refresh_freq - loop_time - 0.01) 
        return decoded_data
    
    def tear_down(self):
        pass
    

def test_quattrolight():

    qt = QuattroOtlight(host='128.95.215.191', refresh_freq=32)
    qt.setup()
    prev_time = time.perf_counter()

    N_reads = 100
    for _  in range(N_reads):
        qt.read_emg()

    post_time = time.perf_counter()
    print(f'it takes {post_time - prev_time} to do {N_reads} reads')
    print(f'mean read time is {(post_time - prev_time) / N_reads}s')
    print(f'sampling freq is then {N_reads/(post_time - prev_time)}')
    qt.tear_down()

def test_quattodummy():
    qt = QuattroDummy()
    qt.setup()
    prev_time = time.perf_counter()

    N_reads = 10
    for _  in range(N_reads):
        qt.read_emg()

    post_time = time.perf_counter()
    print(f'it takes {post_time - prev_time} to do {N_reads} reads')
    print(f'mean read time is {(post_time - prev_time) / N_reads}s')
    print(f'sampling freq is then {N_reads/(post_time - prev_time)}')
    qt.tear_down()

def test_quottro_functions():
    path = 'C:\\Users\\aolab\\'
    name = 'my_test'
    version = '_v1'
    data_output = path + name + version + '.buffer'
    emg_output = path + name + version + '.csv'
    qt = Quattrocento(emg_output=emg_output, data_output=data_output)
    qt.setup()
    qt.teardown()
    qt.setup()
    
    timestamps = []
    prior_time = time.perf_counter()
    for _ in range(1):
        current_time  = time.perf_counter()
        emg = qt.stream_transform()
        timestamps.append(current_time - prior_time)
        prior_time = current_time

    qt.teardown()
    print(emg)



if __name__ == "__main__":

    test_quattrolight()

    #test_quottro_functions()


# Generated by Django 2.2.1 on 2019-11-12 23:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='feature',
            name='import_path',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='system',
            name='input_path',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='system',
            name='processor_path',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='import_path',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='taskentry',
            name='entry_name',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='visible',
            field=models.BooleanField(blank=True, default=True, null=True),
        ),
        migrations.CreateModel(
            name='TaskEntryCollection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(default='', max_length=200)),
                ('entries', models.ManyToManyField(to='tracker.TaskEntry')),
            ],
        ),
    ]



import time
import traceback
import numpy as np
import zmq
from msgpack import loads
import msgpack
import time
from .surface_no_delay import NoDelaySurfaceGazeMapper, RadialDistortionCamera
from .pupillab_timesync import measure_clock_offset

from riglib.source import DataSourceSystem

class System(DataSourceSystem):
    '''
    
    '''
    dtype = np.dtype((float, (12,)))
    update_freq = 120
    
    def __init__(self, ip="128.95.215.191", port="50020", confidence_threshold=0.0):
        '''
        For eye tracking, need Pupil Capture running in the background (after calibration in Pupil Capture)
        '''
        # define a surface AOI
        
        # open a req port to talk to pupil
        self.ip = ip  # remote ip or localhost
        self.port = port # same as in the pupil remote gui
        
        # # matrix for camera distortion
        camera = RadialDistortionCamera(
            resolution=(1280, 720),
            cam_matrix=[
                [794.3311439869655, 0.0, 633.0104437728625],
                [0.0, 793.5290139393004, 397.36927353414865],
                [0.0, 0.0, 1.0],
            ],
            dist_coefs=[
                [
                    -0.3758628065070806,
                    0.1643326166951343,
                    0.00012182540692089567,
                    0.00013422608638039466,
                    0.03343691733865076,
                    0.08235235770849726,
                    -0.08225804883227375,
                    0.14463365333602152,
                ]
            ],
        )

        self.mapper = NoDelaySurfaceGazeMapper(camera)
        self.mapped_points = []
        self.confidence_threshold = confidence_threshold
    
    def start(self):
        '''
        
        '''
        self.ctx = zmq.Context()
        self.pupil_remote = self.ctx.socket(zmq.REQ)
        # The REQ talks to Pupil remote and receives the session unique IPC SUB PORT
        self.pupil_remote.connect(f'tcp://{self.ip}:{self.port}') #connect to eye tracker

        # rec_name = 'test'
        # req.send_string(f'R {rec_name}') # start recording
        # print('pupillab starting recording in Pupil Capture')

        # # sync pupil internal clock with the local time 
        # local_clock = time.time  # Unix time, less accurate 

        # # Measure clock offset once
        # self.offset = measure_clock_offset(pupil_remote, clock_function=local_clock)
        # print(f"\n Pupillab Clock offset (1 measurement): {self.offset} seconds")

        self.pupil_remote.send_string('SUB_PORT') # Request 'SUB_PORT' for reading data
        sub_port = self.pupil_remote.recv_string()

        # open sub ports to listen to pupil; sub: subport that receives surface data
        self.sub = self.ctx.socket(zmq.SUB) # open a sub port to listen to pupil surface topic
        self.sub.connect(f'tcp://{self.ip}:{sub_port}')
        # self.sub.setsockopt_string(zmq.SUBSCRIBE, f"surfaces.{surface_name}")
        self.sub.subscribe(f"surfaces")  # receive all surface messages
        self.sub.subscribe("gaze")  # receive all gaze messages
        self.sub.subscribe('pupil.0.2d')  # receive all pupil0 messages, right eye
        self.sub.subscribe('pupil.1.2d')  # receive all pupil1 messages, left eye

    def stop(self):
        self.sub.close()
        # req.send_string('r') # stop recording
        # print('pupillab stopped recording in Pupil Capture')
        self.pupil_remote.close()
        self.ctx.term()

    def get(self):
        """
        read in a batch of eye data and retun x, y on surface & pupil diameters for each eye
        """
        gaze = np.full((6,), np.nan)
        left_pupil = np.full((3,), np.nan)
        right_pupil = np.full((3,), np.nan)
        coords = np.hstack([gaze, left_pupil, right_pupil])

        while np.any(np.isnan(coords)):
            if not self.sub.poll(0) == zmq.POLLIN:
                time.sleep(0.001)  # sleep for 1 ms to avoid busy waiting
                continue

            topic, payload = self.sub.recv_multipart(flags=zmq.NOBLOCK)
            message = msgpack.loads(payload, raw=False)

            if message["topic"].startswith("surfaces"):
                self.mapper.update_homography(message["img_to_surf_trans"])

            if message["topic"].startswith("gaze.3d.") and message["confidence"] >= self.confidence_threshold:
                    
                    gaze[:2] = message["norm_pos"]
                    gaze[2:5] = message["gaze_point_3d"]

                    if self.mapper is not None:
                        mapped_gaze = self.mapper.gaze_to_surface(message["norm_pos"])
                        if mapped_gaze is not None:
                            gaze[:2] = np.array(mapped_gaze.norm_x, mapped_gaze.norm_y)

                    gaze[5] = message["timestamp"]

            if message["topic"] == "pupil.1.2d" and message["confidence"] >= self.confidence_threshold:
                left_pupil[:2] = message["norm_pos"]
                left_pupil[2] = float(message["diameter"]) # pupil 1 diamter, left eye, unit: pixel
            
            elif message["topic"] == "pupil.0.2d" and message["confidence"] >= self.confidence_threshold:
                right_pupil[:2] = message["norm_pos"] 
                right_pupil[2] = float(message["diameter"]) # pupil 0 diamter, right eye, unit: pixel

            coords = np.hstack([gaze, left_pupil, right_pupil])
            
        coords = np.expand_dims(coords, axis=0)
        return coords

    
class NoSurfaceTracking(System):

    def __init__(self, ip="128.95.215.191", port="50020", confidence_threshold=0.0):
        '''
        For eye tracking, need Pupil Capture running in the background (after calibration in Pupil Capture)
        '''
        # define a surface AOI
        
        # open a req port to talk to pupil
        self.ip = ip  # remote ip or localhost
        self.port = port # same as in the pupil remote gui
        
        self.mapper = None
        self.mapped_points = []

        self.confidence_threshold = confidence_threshold



import numpy as np

def calculate_square_positions(half_height, window_size, n, square_size):
    '''
    Generate the x,y positions of squares tangent to the window perimeter
    '''

    pixel_to_cm_ratio =  2 * half_height / window_size[1]
    width = window_size[0] * pixel_to_cm_ratio - square_size
    height = window_size[1] * pixel_to_cm_ratio - square_size
    n_rounded = 4 * np.ceil(n / 4) # make spacing always a multiple of 4
    horizontal_spacing = (2 * width) / (n_rounded / 2)
    vertical_spacing = (2 * height) / (n_rounded / 2)

    print(horizontal_spacing, vertical_spacing)

    centers = []
    current_distance = 0

    for _ in range(n):
        if current_distance < width:                    # Top edge
            x = current_distance - width / 2 - square_size / 2
            y = height / 2 - square_size / 2
            print('top', current_distance, x, y)
            current_distance += horizontal_spacing
        elif current_distance < width + height:         # Right edge
            x = width / 2 - square_size / 2
            y = height / 2 - (current_distance - width) - square_size / 2
            print('right', current_distance, x, y)
            current_distance += vertical_spacing
        elif current_distance < 2 * width + height:     # Bottom edge
            x = width / 2 - (current_distance - width - height) - square_size / 2
            y = - height / 2 - square_size / 2
            print('bottom', current_distance, x, y)
            current_distance += horizontal_spacing
        else:                                           # Left edge
            x = - width / 2 - square_size / 2
            y = - height / 2 + (current_distance - 2 * width - height) - square_size / 2
            print('left', current_distance, x, y)
            current_distance += vertical_spacing

        centers.append((x, y))

    return centers